------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--       O C A R I N A . G E N E R A T O R S . B U I L D _ U T I L S        --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                 Copyright (C) 2007, GET-Telecom Paris.                   --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;

with GNAT.Table;
with GNAT.Dynamic_Tables;
with GNAT.OS_Lib;
with GNAT.Expect;
with GNAT.Directory_Operations;

with Namet;
with Output;
with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Entities.Components;
with Ocarina.Generators.Utils;
with Ocarina.Generators.Messages;
with Ocarina.Generators.Properties;
with Ocarina.Generators.Ada_Tree.Nutils;

package body Ocarina.Generators.Build_Utils is

   use GNAT.OS_Lib;
   use GNAT.Directory_Operations;
   use Namet;
   use Output;
   use Ocarina.Nodes;
   use Ocarina.Entities.Components;
   use Ocarina.Generators.Utils;
   use Ocarina.Generators.Messages;
   use Ocarina.Generators.Properties;

   package AAU renames Ocarina.Nutils;
   package ADU renames Ocarina.Generators.Ada_Tree.Nutils;

   package Name_Tables is new GNAT.Dynamic_Tables (Name_Id, Nat, 1, 10, 10);
   --  Provides a flexible Name_Id list

   function Length (T : Name_Tables.Instance) return Int;
   --  Return the length of a name table

   procedure Split_Path
     (P        :     Name_Id;
      L        :     Name_Id;
      Basename : out Name_Id;
      Dirname  : out Name_Id);
   --  Split the brute path 'P' into a base name and a directory name
   --  in the context of the current working directory and the
   --  additional location L. The path is even absolute even relative
   --  to the AADL file directory. All relative path are converted
   --  into absolute paths.

   generic
      --  This generic package is a generic list to store the "build
      --  utils" (makefiles, project files...). It provides accessor
      --  routines to allow a process node to find its corresponding
      --  its corresponding "build util".

      type Build_Util is private;
      --  The type of "build util"

      Id : String;
      --  The Id of the generic table. It MUST be a unique string.

      with procedure Free (T : in out Build_Util);
      --  For deallocation purpose

   package Generic_List is
      --  This package is a generic list to store the "build utils"
      --  (makefiles, project files...). It provides accessor routines
      --  to allow a process node to find its corresponding its
      --  corresponding "build util".

      procedure Set (P : Node_Id; U : Build_Util);
      function Get (P : Node_Id) return Build_Util;

      procedure Free;
      --  Deallocates the table

      procedure Init;
      --  A call to this procedure is NECESSARY after any call to
      --  Free. It is not necessary before the first use of the table.

   end Generic_List;

   ------------
   -- Length --
   ------------

   function Length (T : Name_Tables.Instance) return Int is
   begin
      return Name_Tables.Last (T) - Name_Tables.First + 1;
   end Length;

   ----------------
   -- Split_Path --
   ----------------

   procedure Split_Path
     (P        :     Name_Id;
      L        :     Name_Id;
      Basename : out Name_Id;
      Dirname  : out Name_Id)
   is
   begin
      Get_Name_String (P);

      --  Get the directory of the file

      declare
         Dirname_Str  : constant String := Dir_Name
           (Name_Buffer (1 .. Name_Len));
         Basename_Str : constant String := Base_Name
           (Name_Buffer (1 .. Name_Len));
      begin
         if Dirname_Str = "./" or else
           (Dirname_Str'Length > 2 and then
            Dirname_Str (Dirname_Str'First .. Dirname_Str'First + 2) = "../")
         then
            --  This is a relative path. To convert it to an absolute
            --  path, (1) we put the Working directory of Ocarina, (2)
            --  then the directory of the AADL file (if it is
            --  relative) and (3) finally 'Dirname_Str'.

            if L /= No_Name then
               Get_Name_String (L);

               --  If the file location is relative, take in account
               --  the working directory.

               if Name_Buffer (1 .. 2) = "./" or else
                 Name_Buffer (1 .. 3) = "../"
               then
                  Get_Name_String
                    (Add_Directory_Separator (Working_Directory));
                  Get_Name_String_And_Append (L);
               end if;
            else
               Get_Name_String
                 (Add_Directory_Separator (Working_Directory));
            end if;

            Get_Name_String (Add_Directory_Separator (Name_Find));
            Add_Str_To_Name_Buffer (Dirname_Str);

            --  Simplification

            Set_Str_To_Name_Buffer
              (Normalize_Pathname (Name_Buffer (1 .. Name_Len)));

            Dirname := Add_Directory_Separator (Name_Find);
         else
            Dirname := Add_Directory_Separator
              (Get_String_Name (Dirname_Str));
         end if;

         Basename := Get_String_Name (Basename_Str);
      end;
   end Split_Path;

   ------------------
   -- Generic_List --
   ------------------

   package body Generic_List is

      package Internal_Table is new GNAT.Table (Build_Util, Nat, 1, 10, 10);
      --  The interal table

      function Get_Internal_Name (P : Node_Id) return Name_Id;
      --  For code factorization purpose

      -----------------------
      -- Get_Internal_Name --
      -----------------------

      function Get_Internal_Name (P : Node_Id) return Name_Id is
      begin
         pragma Assert (Is_Process (P));

         Set_Nat_To_Name_Buffer (Nat (P));
         Add_Str_To_Name_Buffer ('%' & Id & '%');
         return Name_Find;
      end Get_Internal_Name;

      ---------
      -- Set --
      ---------

      procedure Set (P : Node_Id; U : Build_Util) is
         I_Name : constant Name_Id := Get_Internal_Name (P);
      begin
         Internal_Table.Append (U);
         Set_Name_Table_Info (I_Name, Internal_Table.Last);
      end Set;

      ---------
      -- Get --
      ---------

      function Get (P : Node_Id) return Build_Util is
         I_Name : constant Name_Id := Get_Internal_Name (P);
         Index  : constant Nat := Get_Name_Table_Info (I_Name);
      begin
         if Index = 0 then
            raise Program_Error with "Try to get a build utils which has"
              & " not been set";
         end if;

         return Internal_Table.Table (Index);
      end Get;

      ----------
      -- Init --
      ----------

      procedure Init is
      begin
         Internal_Table.Init;
      end Init;

      ----------
      -- Free --
      ----------

      procedure Free is
      begin
         for J in Internal_Table.First .. Internal_Table.Last loop
            Free (Internal_Table.Table (J));
         end loop;

         Internal_Table.Free;
         Internal_Table.Init;
      end Free;
   end Generic_List;

   ---------------
   -- Makefiles --
   ---------------

   package body Makefiles is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance    (E : Node_Id);
      procedure Visit_System_Instance       (E : Node_Id);
      procedure Visit_Process_Instance      (E : Node_Id);
      procedure Visit_Thread_Instance       (E : Node_Id);
      procedure Visit_Subprogram_Instance   (E : Node_Id);

      procedure Generate_Architecture_Instance (E : Node_Id);
      procedure Generate_Component_Instance    (E : Node_Id);
      procedure Generate_System_Instance       (E : Node_Id);
      procedure Generate_Process_Instance      (E : Node_Id);

      procedure Build_Architecture_Instance (E : Node_Id);
      procedure Build_Component_Instance    (E : Node_Id);
      procedure Build_System_Instance       (E : Node_Id);
      procedure Build_Process_Instance      (E : Node_Id);

      type Makefile_Rec is record
         Appli_Name         : Name_Id;
         --  The distributed application name

         Node_Name          : Name_Id;
         --  The node name (in lower case)

         Execution_Platform : Supported_Execution_Platform;
         --  The execution platform of the processor the current node
         --  is bound to.

         Transport_Layer    : Supported_Transport_Layers;
         --  The transport layer used by the current node to
         --  communicate with other nodes.

         C_Sources          : Name_Tables.Instance;
         --  The C source files that may implement some subprograms of
         --  the current node (absolute or relative path)

         C_Libraries        : Name_Tables.Instance;
         --  The C libraries that may contain the binary code of some
         --  subprograms of the current node (absolute or relative
         --  path)
      end record;
      --  This structure gathers all the information needed to
      --  generate a makefile for a given node of the distributed
      --  application.

      type Makefile_Type is access all Makefile_Rec;

      procedure Free (M : in out Makefile_Type);
      --  Deallocates the internals of T

      procedure PolyORB_HI_Ada_Makefile  (M : Makefile_Type);
      procedure PolyORB_HI_C_Makefile    (M : Makefile_Type);
      procedure PolyORB_Qos_Ada_Makefile (M : Makefile_Type);
      --  Generate the part of the Makefile that is specific to the
      --  corresponding code generator.

      procedure Ada_C_Command_Line_Flags
        (M      : Makefile_Type;
         L_Flag : Boolean);
      procedure Compile_C_Files (M : Makefile_Type);
      --  Code factorization between different generators

      ----------
      -- Free --
      ----------

      procedure Free (M : in out Makefile_Type) is
         procedure Deallocate is new Ada.Unchecked_Deallocation
           (Makefile_Rec, Makefile_Type);
      begin
         Name_Tables.Free (M.all.C_Sources);
         Name_Tables.Free (M.all.C_Libraries);
         Deallocate (M);
      end Free;

      package Makefiles is new Generic_List
        (Makefile_Type, "Makefile_List", Free);
      --  The list of all the makefile structures

      -----------
      -- Reset --
      -----------

      procedure Reset is
      begin
         Makefiles.Free;
      end Reset;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when CC_Subprogram =>
               Visit_Subprogram_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         S  : constant Node_Id := Parent_Subcomponent (E);
         A  : constant Node_Id := Parent_Component (Parent_Subcomponent (E));
         M  : constant Makefile_Type := new Makefile_Rec;
         SC : Node_Id;
      begin
         --  Associates the Makefile structure to the process
         --  instance. Keep in mind that it is important to use
         --  accessed here because all the visited threds and
         --  subprgrams will getch this access to update the
         --  corresponding structure.

         Makefiles.Set (E, M);

         M.Appli_Name := Normalize_Name (Name (Identifier (A)));
         M.Node_Name  := Normalize_Name (Name (Identifier (S)));

         --  Get the execution platform of the processor this node is
         --  bound to.

         M.Execution_Platform := Get_Execution_Platform
           (Get_Bound_Processor (E));

         --  Get the transport layer used by this node. It is
         --  important to ensure that the Namings package visitors
         --  have already been executed since they perform all
         --  consistency checks and bind a node to its transport
         --  layer.

         M.Transport_Layer := Fetch_Transport_Layer (E);

         --  Initialize the lists

         Name_Tables.Init (M.C_Sources);
         Name_Tables.Init (M.C_Libraries);

         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            SC := First_Node (Subcomponents (E));

            while Present (SC) loop
               --  Visit the corresponding instance of SC

               Visit (Corresponding_Instance (SC));

               SC := Next_Node (SC);
            end loop;
         end if;
      end Visit_Process_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         Call_Seq : Node_Id;
         Spg_Call : Node_Id;
      begin
         --  Visit all the call sequences of the thread

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;
      end Visit_Thread_Instance;

      -------------------------------
      -- Visit_Subprogram_Instance --
      -------------------------------

      procedure Visit_Subprogram_Instance (E : Node_Id) is
         Parent_Process  : constant Node_Id :=
           Corresponding_Instance (Get_Container_Process (E));
         M               : constant Makefile_Type :=
           Makefiles.Get (Parent_Process);
         Subprogram_Kind : constant Supported_Subprogram_Kind :=
           Get_Subprogram_Kind (E);
         Source_Files    : constant Name_Array :=
           Get_Subprogram_Source_Text (E);
         Source_Basename : Name_Id;
         Source_Dirname  : Name_Id;
         S_Name          : Name_Id;
         Call_Seq        : Node_Id;
         Spg_Call        : Node_Id;
      begin
         --  Only C subprogram influence the structure of the
         --  generated makefile.

         case Subprogram_Kind is
            when Subprogram_Opaque_C =>
               --  If the subprogram is implemented by C source files,
               --  add the files to the C_Files list of the makefile
               --  structure. If the subprogram is implemented by a C
               --  library, add the files to the C_Libraries list of
               --  the makefile structure.

               for J in Source_Files'Range loop
                  --  Ensure the source is added only once per node

                  Get_Name_String (Source_Files (J));
                  Get_Name_String_And_Append (M.Node_Name);
                  Add_Str_To_Name_Buffer ("%source_text%");
                  S_Name := Name_Find;

                  if Get_Name_Table_Info (S_Name) = 0 then
                     Set_Name_Table_Info (S_Name, 1);

                     Get_Name_String (Source_Files (J));

                     Split_Path
                       (Source_Files (J),
                        Loc (E).Dir,
                        Source_Basename,
                        Source_Dirname);

                     Get_Name_String (Source_Basename);

                     if Name_Buffer (Name_Len - 1 .. Name_Len) = ".o"
                       or else Name_Buffer (Name_Len - 1 .. Name_Len) = ".a"
                     then
                        --  Library names MUST begin with "lib"

                        if Name_Buffer (Name_Len - 1 .. Name_Len) = ".a"
                          and then (Name_Len <= 5
                                    or else Name_Buffer (1 .. 3) /= "lib")
                        then
                           Display_Error
                             ("Invalid library name"
                              & Name_Buffer (1 .. Name_Len),
                              Fatal => True);
                        end if;

                        Get_Name_String (Source_Dirname);
                        Get_Name_String_And_Append (Source_Basename);

                        Name_Tables.Append (M.C_Libraries, Name_Find);
                     elsif Name_Buffer (Name_Len - 1 .. Name_Len) = ".c" then
                        Get_Name_String (Source_Dirname);
                        Get_Name_String_And_Append (Source_Basename);

                        Name_Tables.Append (M.C_Sources, Name_Find);
                     end if;
                  end if;
               end loop;

            when others =>
               null;
         end case;

         --  Visit all the call sequences of the subprogram

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;
      end Visit_Subprogram_Instance;

      --------------
      -- Generate --
      --------------

      procedure Generate (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Generate_Architecture_Instance (E);

            when K_Component_Instance =>
               Generate_Component_Instance (E);

            when others =>
               null;
         end case;
      end Generate;

      ------------------------------------
      -- Generate_Architecture_Instance --
      ------------------------------------

      procedure Generate_Architecture_Instance (E : Node_Id) is
      begin
         Generate (Root_System (E));
      end Generate_Architecture_Instance;

      ---------------------------------
      -- Generate_Component_Instance --
      ---------------------------------

      procedure Generate_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Generate_System_Instance (E);

            when CC_Process =>
               Generate_Process_Instance (E);

            when others =>
               null;
         end case;
      end Generate_Component_Instance;

      ------------------------------
      -- Generate_System_Instance --
      ------------------------------

      procedure Generate_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Generate the makefiles of all process subcomponents

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));

            while Present (S) loop
               Generate (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;
      end Generate_System_Instance;

      -------------------------------
      -- Generate_Process_Instance --
      -------------------------------

      procedure Generate_Process_Instance (E : Node_Id) is
         M  : constant Makefile_Type := Makefiles.Get (E);
         Fd : File_Descriptor;
      begin
         --  Enter the directories

         Enter_Directory (M.Appli_Name);
         Enter_Directory (M.Node_Name);

         --  Create the file

         Fd := Create_File ("Makefile", Text);

         if Fd = Invalid_FD then
            raise Program_Error;
         end if;

         --  Setting the output

         Set_Output (Fd);

         Write_Line ("#####################################################");
         Write_Line ("# This Makefile has been generated automatically    #");
         Write_Line ("# by the Ocarina AADL toolsuite.                    #");
         Write_Line ("# Do not edit this file since all your changes will #");
         Write_Line ("# be overriddedn at the next code generation.       #");
         Write_Line ("#####################################################");
         Write_Eol;

         Write_Str ("# Distributed application name : ");
         Write_Name (M.Appli_Name);
         Write_Eol;

         Write_Str ("# Node name                    : ");
         Write_Name (M.Node_Name);
         Write_Eol;

         Write_Line ("#  Execution platform          : "
                     & M.Execution_Platform'Img);
         Write_Line ("#  Transport layer             : "
                     & M.Transport_Layer'Img);

         Write_Eol;

         case Current_Generator_Kind is
            when PolyORB_HI_Ada =>
               PolyORB_HI_Ada_Makefile (M);
            when PolyORB_QoS_Ada =>
               PolyORB_QoS_Ada_Makefile (M);
            when PolyORB_HI_C =>
               PolyORB_HI_C_Makefile (M);
            when others =>
               raise Program_Error;
         end case;

         --  Close the file

         Close (Fd);
         Set_Standard_Output;

         --  Leave the directories

         Leave_Directory;
         Leave_Directory;
      end Generate_Process_Instance;

      -----------------------------
      -- PolyORB_HI_Ada_Makefile --
      -----------------------------

      procedure PolyORB_HI_Ada_Makefile (M : Makefile_Type) is
         --  N : Name_Id;
      begin
         --  Variable definitions

         --  Project file

         Write_Str ("PROJECT_FILE = ");
         Write_Name (M.Node_Name);
         Write_Line (".gpr");

         --  Compiler

         case M.Execution_Platform is
            when Platform_Native =>
               Write_Line ("GNATMAKE = gnatmake");
               Write_Line ("GNAT = gnat");
               Write_Line ("GCC = gcc");
               Write_Line ("TARGET = NATIVE");

            when Platform_LEON =>
               Write_Line ("GNATMAKE = sparc-elf-gnatmake");
               Write_Line ("GNAT = sparc-elf-gnat");
               Write_Line ("GCC = sparc-elf-gcc");
               Write_Line ("TARGET = LEON");

            when PLATFORM_ERC_32 =>
               Write_Line ("GNATMAKE = erc32-elf-gnatmake");
               Write_Line ("GNAT = erc32-elf-gnat");
               Write_Line ("GCC = erc32-elf-gcc");
               Write_Line ("TARGET = ERC32");

            when others =>
               raise Program_Error with "Unknown execution platform";
         end case;

         Write_Eol;

         --  The 'all' target

         Write_Str ("all:");

         --  If there are C files to be compiled add a dependency on
         --  these files

         if Length (M.C_Sources) > 0 then
            Write_Str (" compile-c-files");
         end if;

         Write_Eol;

         Write_Char (ASCII.HT);
         Write_Str ("$(GNATMAKE) -P$(PROJECT_FILE) -XTARGET=$(TARGET)");

         --  If there are C source or C libraries, there will be more
         --  options.

         Ada_C_Command_Line_Flags (M, True);
         Write_Eol;
         Write_Eol;

         --  Compile the C files, if any

         Compile_C_Files (M);
         Write_Eol;
      end PolyORB_HI_Ada_Makefile;

      ---------------------------
      -- PolyORB_HI_C_Makefile --
      ---------------------------

      procedure PolyORB_HI_C_Makefile (M : Makefile_Type) is
      begin
         Write_Str ("BINARY = ");
         Write_Name (M.Node_Name);
         Write_Eol;

         --  Variable definitions
         case M.Execution_Platform is
            when Platform_Native =>
               Write_Line ("CC = gcc");
               Write_Line ("TARGET = NATIVE");

            when Platform_LEON =>
               Write_Line ("CC = sparc-elf-gcc");
               Write_Line ("TARGET = LEON");

            when PLATFORM_ERC_32 =>
               Write_Line ("CC = erc32-elf-gcc");
               Write_Line ("TARGET = ERC32");

            when others =>
               raise Program_Error with "Unknown execution platform";
         end case;

         Write_Eol;
         Write_Line ("include $(POLYORB_HI_PATH)/src/Makefile.common");
         Write_Eol;
         Write_Str ("all: compile-node");
         --  The 'all' target

         Write_Eol;
      end PolyORB_HI_C_Makefile;

      ------------------------------
      -- PolyORB_Qos_Ada_Makefile --
      ------------------------------

      procedure PolyORB_Qos_Ada_Makefile (M : Makefile_Type) is
      begin
         --  Project file

         Write_Str ("PROJECT_FILE = ");
         Write_Name (M.Node_Name);
         Write_Line (".gpr");

         --  Variable definitions

         Write_Line ("GNATMAKE = gnatmake");
         Write_Line ("GNAT = gnat");
         Write_Line ("GCC = gcc");

         Write_Eol;

         --  The 'all' target

         Write_Str ("all: ");

         --  If there are C files to be compiled add a dependency on
         --  these files

         if Length (M.C_Sources) > 0 then
            Write_Str (" compile-c-files");
         end if;

         Write_Eol;

         Write_Str
           (ASCII.HT
            & "$(GNAT) make -P$(PROJECT_FILE) `polyorb-config --cflags`\"
            & ASCII.LF
            & ASCII.HT
            & "   -largs `polyorb-config --libs` ");

         --  If there are C source or C libraries, there will be more
         --  options.

         Ada_C_Command_Line_Flags (M, False);
         Write_Eol;

         --  Override the polyorb-parameters-partition.adb object file

         if Is_Regular_File ("polyorb-parameters-partition.adb") then
            --  Compile polyorb-parameters-partition.adb

            Write_Line
              (ASCII.HT &
               "$(GNAT) compile -gnatwe -a polyorb-parameters-partition.adb\");
            Write_Line (ASCII.HT & "   `polyorb-config --cflags`;\");

            --  Bind

            Write_Line
              (ASCII.HT
               & "$(GNAT) bind -P$(PROJECT_FILE) `polyorb-config --cflags`\");
            Write_Str (ASCII.HT & "   ");
            Write_Name (M.Node_Name);
            Write_Line (".ali");

            --  Link

            Write_Str (ASCII.HT & "$(GNAT) link -P$(PROJECT_FILE) ");
            Write_Name (M.Node_Name);
            Write_Line (".ali\");
            Write_Str (ASCII.HT & "   `polyorb-config --libs` ");

            --  If there are C source or C libraries, there will be more
            --  options.

            Ada_C_Command_Line_Flags (M, False);
            Write_Eol;
         end if;

         Write_Eol;

         --  Compile the C files, if any

         Compile_C_Files (M);
         Write_Eol;
      end PolyORB_Qos_Ada_Makefile;

      ------------------------------
      -- Ada_C_Command_Line_Flags --
      ------------------------------

      procedure Ada_C_Command_Line_Flags
        (M      : Makefile_Type;
         L_Flag : Boolean)
      is
      begin
         if L_Flag and then
           (Length (M.C_Sources) > 0 or else Length (M.C_Libraries) > 0)
         then
            Write_Line (" -largs \");
            Write_Str (ASCII.HT & "   ");
         end if;

         --  In case of C source files, we add the correspodning .o
         --  files.

         if Length (M.C_Sources) > 0 then
            for J in Name_Tables.First .. Name_Tables.Last (M.C_Sources) loop
               Get_Name_String (M.C_Sources.Table (J));
               Set_Str_To_Name_Buffer
                 (Base_Name
                  (Name_Buffer
                   (1 .. Name_Len)));
               Name_Buffer (Name_Len) := 'o';
               Write_Name (Name_Find);

               exit when J = Name_Tables.Last (M.C_Sources);

               Write_Line (" \");
               Write_Str (ASCII.HT & "   ");
            end loop;
         end if;

         --  In case of C libraries or object, we add the
         --  corresponding option.

         if Length (M.C_Libraries) > 0 then
            Write_Line (" \");
            Write_Str (ASCII.HT & "   ");

            for J in Name_Tables.First .. Name_Tables.Last (M.C_Libraries) loop
               Get_Name_String (M.C_Libraries.Table (J));

               --  Some tests

               declare
                  Is_Object : constant Boolean := Name_Buffer (Name_Len) = 'o';
                  Dirname   : constant String := Dir_Name
                    (Name_Buffer (1 .. Name_Len));
                  Basename  : constant String :=
                    Base_Name (Name_Buffer (1 .. Name_Len));
               begin
                  if Is_Object then
                     Write_Name (M.C_Libraries.Table (J));
                  else
                     Write_Str ("-L" & Dirname & ' ');
                     Write_Str ("-l");
                     Write_Str
                       (Basename (Basename'First + 3 .. Basename'Last));
                  end if;
               end;

               exit when J = Name_Tables.Last (M.C_Libraries);

               Write_Line (" \");
               Write_Str (ASCII.HT & "   ");
            end loop;
         end if;
      end Ada_C_Command_Line_Flags;

      ---------------------
      -- Compile_C_Files --
      ---------------------

      procedure Compile_C_Files (M : Makefile_Type) is
      begin
         if Length (M.C_Sources) > 0 then
            Write_Line ("compile-c-files:");

            for J in Name_Tables.First .. Name_Tables.Last (M.C_Sources) loop
               declare
                  O_File : Name_Id;
               begin
                  Get_Name_String (M.C_Sources.Table (J));
                  Name_Buffer (Name_Len) := 'o';
                  Set_Str_To_Name_Buffer
                    (Base_Name (Name_Buffer (1 .. Name_Len)));
                  O_File := Name_Find;

                  Write_Char (ASCII.HT);
                  Write_Str ("$(GCC) -c ");
                  Write_Name (M.C_Sources.Table (J));
                  Write_Str (" -o ");
                  Write_Name (O_File);
                  Write_Eol;
               end;
            end loop;
         end if;
      end Compile_C_Files;

      -----------
      -- Build --
      -----------

      procedure Build (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Build_Architecture_Instance (E);

            when K_Component_Instance =>
               Build_Component_Instance (E);

            when others =>
               null;
         end case;
      end Build;

      ---------------------------------
      -- Build_Architecture_Instance --
      ---------------------------------

      procedure Build_Architecture_Instance (E : Node_Id) is
      begin
         Build (Root_System (E));
      end Build_Architecture_Instance;

      ------------------------------
      -- Build_Component_Instance --
      ------------------------------

      procedure Build_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Build_System_Instance (E);

            when CC_Process =>
               Build_Process_Instance (E);

            when others =>
               null;
         end case;
      end Build_Component_Instance;

      ---------------------------
      -- Build_System_Instance --
      ---------------------------

      procedure Build_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Build all process subcomponents

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));

            while Present (S) loop
               Build (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;
      end Build_System_Instance;

      ----------------------------
      -- Build_Process_Instance --
      ----------------------------

      procedure Build_Process_Instance (E : Node_Id) is
         use GNAT.Expect;

         M      : constant Makefile_Type := Makefiles.Get (E);
         Fd     : Process_Descriptor;
         Result : Expect_Match;
         Status : Integer := 0;
         Args   : Argument_List (1 .. 0);
      begin
         --  Enter the directories

         Enter_Directory (M.Appli_Name);
         Enter_Directory (M.Node_Name);

         --  Invoke the 'make' command

         Non_Blocking_Spawn
           (Descriptor  => Fd,
            Command     => "make",
            Args        => Args,
            Buffer_Size => 4096,
            Err_To_Out  => True);

         Add_Filter (Fd, Trace_Filter'Access, GNAT.Expect.Output);
         Add_Filter (Fd, Trace_Filter'Access, GNAT.Expect.Died);

         --  Wait until the command achieves its execution

         begin
            Expect (Fd, Result, "", -1);

            --  This call never match, it is only used to wait until
            --  Ocarina1 terminates.

         exception
            when GNAT.Expect.Process_Died =>
               Close (Fd, Status);

               if Status /= 0 then
                  Display_Error ("make died unexpectedly with code: "
                                 & Status'Img,
                                 Fatal => True);
               else
                  pragma Debug (Display_Debug_Message
                                ("make finished normally", Force => True));
                  null;
               end if;
         end;

         --  Leave the directories

         Leave_Directory;
         Leave_Directory;
      end Build_Process_Instance;

   end Makefiles;

   -----------------------
   -- Ada_Project_Files --
   -----------------------

   package body Ada_Project_Files is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance    (E : Node_Id);
      procedure Visit_System_Instance       (E : Node_Id);
      procedure Visit_Process_Instance      (E : Node_Id);
      procedure Visit_Thread_Instance       (E : Node_Id);
      procedure Visit_Subprogram_Instance   (E : Node_Id);

      procedure Generate_Architecture_Instance (E : Node_Id);
      procedure Generate_Component_Instance    (E : Node_Id);
      procedure Generate_System_Instance       (E : Node_Id);
      procedure Generate_Process_Instance      (E : Node_Id);

      type Ada_Project_File_Rec is record
         Appli_Name         : Name_Id;
         --  The distributed application name

         Node_Name          : Name_Id;
         --  The node name (in lower case)

         Is_Server          : Boolean;
         --  True of the process has IN ports

         Execution_Platform : Supported_Execution_Platform;
         --  The execution platform of the processor the current node
         --  is bound to.

         Transport_Layer    : Supported_Transport_Layers;
         --  The transport layer used by the current node to
         --  communicate with other nodes.

         Spec_Names        : Name_Tables.Instance;
         Custom_Spec_Names : Name_Tables.Instance;
         --  USER Ada specs with custom names. For each index J,
         --  Spec_Names (J) is the Ada spec name and Custom_Spec_Names
         --  (J) is the file name containing the spec.

         Body_Names        : Name_Tables.Instance;
         Custom_Body_Names : Name_Tables.Instance;
         --  USER Ada bodies with custom names. For each index J,
         --  Body_Names (J) is the Ada body name and Custom_Body_Names
         --  (J) is the file name containing the body.

         User_Source_Dirs : Name_Tables.Instance;
         --  Directories of the source files provided by the user
      end record;
      --  This structure gathers all the information needed to
      --  generate an Ada project file for a given node of the
      --  distributed application.

      type Ada_Project_File_Type is access all Ada_Project_File_Rec;

      procedure Free (P : in out Ada_Project_File_Type);
      --  Deallocates the internals of T

      procedure PolyORB_HI_Ada_Ada_Project_File  (P : Ada_Project_File_Type);
      procedure PolyORB_HI_C_Ada_Project_File    (P : Ada_Project_File_Type);
      procedure PolyORB_Qos_Ada_Ada_Project_File (P : Ada_Project_File_Type);
      --  Generate the part of the Ada project file that is specific
      --  to the corresponding code generator.

      ----------
      -- Free --
      ----------

      procedure Free (P : in out Ada_Project_File_Type) is
         procedure Deallocate is new Ada.Unchecked_Deallocation
           (Ada_Project_File_Rec, Ada_Project_File_Type);
      begin
         --  Deallocate internal tables

         Name_Tables.Free (P.all.Spec_Names);
         Name_Tables.Free (P.all.Custom_Spec_Names);
         Name_Tables.Free (P.all.Body_Names);
         Name_Tables.Free (P.all.Custom_Body_Names);
         Name_Tables.Free (P.all.User_Source_Dirs);

         Deallocate (P);
      end Free;

      package Ada_Project_Files is new Generic_List
        (Ada_Project_File_Type, "Ada_Project_File_List", Free);
      --  The list of all the makefile structures

      -----------
      -- Reset --
      -----------

      procedure Reset is
      begin
         Ada_Project_Files.Free;
      end Reset;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when CC_Subprogram =>
               Visit_Subprogram_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         S  : constant Node_Id := Parent_Subcomponent (E);
         A  : constant Node_Id := Parent_Component (Parent_Subcomponent (E));
         P  : constant Ada_Project_File_Type := new Ada_Project_File_Rec;
         SC : Node_Id;
      begin
         --  Associates the Ada project file structure to the process
         --  instance. Keep in mind that it is important to use
         --  accessed here because all the visited threds and
         --  subprgrams will getch this access to update the
         --  corresponding structure.

         Ada_Project_Files.Set (E, P);

         P.Appli_Name := Normalize_Name (Name (Identifier (A)));
         P.Node_Name  := Normalize_Name (Name (Identifier (S)));

         P.Is_Server := Has_In_Ports (E);

         --  Get the execution platform of the processor this node is
         --  bound to.

         P.Execution_Platform := Get_Execution_Platform
           (Get_Bound_Processor (E));

         --  Get the transport layer used by this node. It is
         --  important to ensure that the Namings package visitors
         --  have already been executed since they perform all
         --  consistency checks and bind a node to its transport
         --  layer.

         P.Transport_Layer := Fetch_Transport_Layer (E);

         --  Initialize the lists

         Name_Tables.Init (P.Spec_Names);
         Name_Tables.Init (P.Custom_Spec_Names);

         Name_Tables.Init (P.Body_Names);
         Name_Tables.Init (P.Custom_Body_Names);

         Name_Tables.Init (P.User_Source_Dirs);

         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            SC := First_Node (Subcomponents (E));

            while Present (SC) loop
               --  Visit the corresponding instance of SC

               Visit (Corresponding_Instance (SC));

               SC := Next_Node (SC);
            end loop;
         end if;
      end Visit_Process_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         Call_Seq : Node_Id;
         Spg_Call : Node_Id;
      begin
         --  Visit all the call sequences of the thread

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;
      end Visit_Thread_Instance;

      -------------------------------
      -- Visit_Subprogram_Instance --
      -------------------------------

      procedure Visit_Subprogram_Instance (E : Node_Id) is
         Parent_Process  : constant Node_Id :=
           Corresponding_Instance (Get_Container_Process (E));
         P               : constant Ada_Project_File_Type :=
           Ada_Project_Files.Get (Parent_Process);
         Subprogram_Kind : constant Supported_Subprogram_Kind :=
           Get_Subprogram_Kind (E);
         Source_Name     : constant Name_Id :=
           Get_Subprogram_Source_Name (E);
         Source_Files    : constant Name_Array :=
           Get_Subprogram_Source_Text (E);
         Conv_Base_Name  : Name_Id;
         Custom_Name     : Name_Id;
         Suffix          : String (1 .. 4);
         Source_Dirname  : Name_Id;
         Source_Basename : Name_Id;
         Binding_Key     : constant String := "%user_src_dir%";
         Call_Seq        : Node_Id;
         Spg_Call        : Node_Id;
      begin
         --  Only Ada subprograms may influence the structure of the
         --  generated project files.

         case Subprogram_Kind is
            when Subprogram_Opaque_Ada_95 | Subprogram_Hybrid_Ada_95 =>

               Conv_Base_Name  := ADU.Conventional_Base_Name
                 (ADU.Unit_Name (Source_Name));

               --  Ensure the user gives at most 2 sources files (a
               --  spec and a body).

               if Source_Files'Length > 2 then
                  Display_Located_Error
                    (Loc (E),
                     "More than 2 source files for an Ada subprogram",
                     Fatal => True);
               end if;

               if Source_Files'Length = 0 then
                  --  This means that the user did not provide source
                  --  file names for the Ada subprograms. This means
                  --  also that the corresponding source files have
                  --  conventional names and are located at the same
                  --  directory as the AADL file.

                  Split_Path
                    (Conv_Base_Name,
                     Loc (E).Dir,
                     Source_Basename,
                     Source_Dirname);

                  Set_Str_To_Name_Buffer (Binding_Key);
                  Get_Name_String_And_Append (Source_Dirname);
                  Get_Name_String_And_Append (P.Node_Name);

                  if Get_Name_Table_Byte (Name_Find) = 0 then
                     Name_Tables.Append
                       (P.User_Source_Dirs,
                        Source_Dirname);
                     Set_Name_Table_Byte (Name_Find, 1);
                  end if;
               else
                  for J in Source_Files'Range loop
                     Split_Path
                       (Source_Files (J),
                        Loc (E).Dir,
                        Source_Basename,
                        Source_Dirname);

                     --  Add the directory to the user directory list
                     --  (if it has not been added yet)

                     Set_Str_To_Name_Buffer (Binding_Key);
                     Get_Name_String_And_Append (Source_Dirname);
                     Get_Name_String_And_Append (P.Node_Name);

                     if Get_Name_Table_Byte (Name_Find) = 0 then
                        Name_Tables.Append
                          (P.User_Source_Dirs,
                           Source_Dirname);
                        Set_Name_Table_Byte (Name_Find, 1);
                     end if;

                     Get_Name_String (Source_Basename);

                     --  The .ad[bs] consumes 4 characters from to
                     --  total file name. The user must give at leas
                     --  on character base name.

                     if Name_Len < 5 then
                        Display_Located_Error
                          (Loc (E),
                           "Incorrect text file name",
                           Fatal => True);
                     end if;

                     Suffix := Name_Buffer (Name_Len - 3 .. Name_Len);

                     Custom_Name := Name_Find;

                     if Suffix = ".ads" then
                        if Custom_Name /= Conv_Base_Name then
                           --  Add a custom Spec clause

                           Name_Tables.Append
                             (P.Spec_Names, ADU.Unit_Name (Source_Name));
                           Name_Tables.Append
                             (P.Custom_Spec_Names, Custom_Name);
                        end if;
                     elsif Suffix = ".adb" then
                        if Custom_Name /= Conv_Base_Name then
                           --  Add a custom Body clause

                           Name_Tables.Append
                             (P.Body_Names, ADU.Unit_Name (Source_Name));
                           Name_Tables.Append
                             (P.Custom_Body_Names, Custom_Name);
                        end if;
                     else
                        Display_Located_Error
                          (Loc (E),
                           "Unknown suffix for Ada file name: """
                           & Suffix
                           & """",
                           Fatal => True);
                     end if;
                  end loop;
               end if;
            when others =>
               null;
         end case;

         --  Visit all the call sequences of the subprogram

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;
      end Visit_Subprogram_Instance;

      --------------
      -- Generate --
      --------------

      procedure Generate (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Generate_Architecture_Instance (E);

            when K_Component_Instance =>
               Generate_Component_Instance (E);

            when others =>
               null;
         end case;
      end Generate;

      ------------------------------------
      -- Generate_Architecture_Instance --
      ------------------------------------

      procedure Generate_Architecture_Instance (E : Node_Id) is
      begin
         Generate (Root_System (E));
      end Generate_Architecture_Instance;

      ---------------------------------
      -- Generate_Component_Instance --
      ---------------------------------

      procedure Generate_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Generate_System_Instance (E);

            when CC_Process =>
               Generate_Process_Instance (E);

            when others =>
               null;
         end case;
      end Generate_Component_Instance;

      ------------------------------
      -- Generate_System_Instance --
      ------------------------------

      procedure Generate_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Generate the project files of all process subcomponents

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));

            while Present (S) loop
               Generate (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;
      end Generate_System_Instance;

      -------------------------------
      -- Generate_Process_Instance --
      -------------------------------

      procedure Generate_Process_Instance (E : Node_Id) is
         P  : constant Ada_Project_File_Type := Ada_Project_Files.Get (E);
         Fd : File_Descriptor;
      begin
         --  Enter the directories

         Enter_Directory (P.Appli_Name);
         Enter_Directory (P.Node_Name);

         --  Create the file

         Get_Name_String (P.Node_Name);
         Fd := Create_File (Name_Buffer (1 .. Name_Len) & ".gpr", Text);

         if Fd = Invalid_FD then
            raise Program_Error;
         end if;

         --  Setting the output

         Set_Output (Fd);

         Write_Line
           ("--------------------------------------------------------");
         Write_Line
           ("-- This project file has been generated automatically --");
         Write_Line
           ("-- by the Ocarina AADL toolsuite.                     --");
         Write_Line
           ("-- Do not edit this file since all your changes will  --");
         Write_Line
           ("-- be overriddedn at the next code generation.        --");
         Write_Line
           ("--------------------------------------------------------");
         Write_Eol;

         Write_Str ("--  Distributed application name : ");
         Write_Name (P.Appli_Name);
         Write_Eol;

         Write_Str ("--  Node name                    : ");
         Write_Name (P.Node_Name);
         Write_Eol;

         Write_Line ("--  Execution platform           : "
                     & P.Execution_Platform'Img);
         Write_Line ("--  Transport layer              : "
                     & P.Transport_Layer'Img);

         Write_Eol;

         case Current_Generator_Kind is
            when PolyORB_HI_Ada =>
               PolyORB_HI_Ada_Ada_Project_File (P);
            when PolyORB_QoS_Ada =>
               PolyORB_QoS_Ada_Ada_Project_File (P);
            when PolyORB_HI_C =>
               PolyORB_HI_C_Ada_Project_File (P);
            when others =>
               raise Program_Error;
         end case;

         --  Close the file

         Close (Fd);
         Set_Standard_Output;

         --  Leave the directories

         Leave_Directory;
         Leave_Directory;
      end Generate_Process_Instance;

      -------------------------------------
      -- PolyORB_HI_Ada_Ada_Project_File --
      -------------------------------------

      procedure PolyORB_HI_Ada_Ada_Project_File (P : Ada_Project_File_Type) is
      begin
         --  If the transport layer is SpaceWire, add the dependency
         --  to the project file gathering the SpaceWire driver.

         if P.Transport_Layer = Transport_SpaceWire then
            Write_Line ("with ""mts_ada_binding.gpr"";");
            Write_Eol;
         end if;

         Write_Str ("project ");
         Write_Name (P.Node_Name);
         Write_Line (" extends ""polyorb_hi"" is");
         Increment_Indentation;

         --  The source directory list

         Write_Indentation;
         Write_Line ("for Source_Dirs use");
         Increment_Indentation;

         Write_Indentation (-1);
         Write_Line ("(""."",");
         Write_Indentation;
         Write_Line ("""../common"",");

         --  The PolyORB-HI source directory is stored in the
         --  environment variable $POLYORB_HI_PATH

         declare
            PO_HI_Src_Dir : String_Access := Getenv ("POLYORB_HI_PATH");
         begin
            if PO_HI_Src_Dir.all = "" then
               Free (PO_HI_Src_Dir);
               Display_Error
                 ("The environment variable ""POLYORB_HI_PATH"" must"
                  & " be set and must point to the PolyORB-HI main"
                  & " directory",
                  Fatal => True);
            else
               Write_Indentation;
               Write_Str ("""" & PO_HI_Src_Dir.all & "/src""");
               Free (PO_HI_Src_Dir);
            end if;
         end;

         --  The user provided source dirs

         if Length (P.User_Source_Dirs) > 0 then
            Write_Line (",");

            for J in
              Name_Tables.First .. Name_Tables.Last (P.User_Source_Dirs) loop
               Write_Indentation;
               Write_Char ('"');
               Write_Name (P.User_Source_Dirs.Table (J));
               Write_Char ('"');

               exit when J = Name_Tables.Last (P.User_Source_Dirs);

               Write_Line (",");
            end loop;
         end if;

         Write_Line (");");

         Decrement_Indentation;

         --  The main subprogram name

         Write_Eol;
         Write_Indentation;
         Write_Str ("for Main use (""");
         Write_Name (P.Node_Name);
         Write_Line (".adb"");");

         --  The custom file names

         Write_Eol;
         Write_Indentation;
         Write_Line ("Package Naming is");
         Increment_Indentation;

         Write_Eol;
         Write_Indentation;
         Write_Line ("--  Custom generated file names");
         Write_Eol;

         Get_Name_String (P.Node_Name);
         Add_Str_To_Name_Buffer ("_deployment.ads");

         if Is_Regular_File (Name_Buffer (1 .. Name_Len)) then
            Write_Indentation;
            Write_Str ("for Specification (""Deployment"")  use """);
            Write_Name (P.Node_Name);
            Write_Line ("_deployment.ads"";");
         end if;

         Get_Name_String (P.Node_Name);
         Add_Str_To_Name_Buffer ("_naming.ads");

         if Is_Regular_File (Name_Buffer (1 .. Name_Len)) then
            Write_Indentation;
            Write_Str ("for Specification (""Naming"")      use """);
            Write_Name (P.Node_Name);
            Write_Line ("_naming.ads"";");
         end if;

         Get_Name_String (P.Node_Name);
         Add_Str_To_Name_Buffer ("_types.ads");

         if Is_Regular_File (Name_Buffer (1 .. Name_Len)) then
            Write_Indentation;
            Write_Str ("for Specification (""Types"")       use """);
            Write_Name (P.Node_Name);
            Write_Line ("_types.ads"";");
         end if;

         Get_Name_String (P.Node_Name);
         Add_Str_To_Name_Buffer ("_types.adb");

         if Is_Regular_File (Name_Buffer (1 .. Name_Len)) then
            Write_Indentation;
            Write_Str ("for Body          (""Types"")       use """);
            Write_Name (P.Node_Name);
            Write_Line ("_types.adb"";");
         end if;

         Get_Name_String (P.Node_Name);
         Add_Str_To_Name_Buffer ("_subprograms.ads");

         if Is_Regular_File (Name_Buffer (1 .. Name_Len)) then
            Write_Indentation;
            Write_Str ("for Specification (""Subprograms"") use """);
            Write_Name (P.Node_Name);
            Write_Line ("_subprograms.ads"";");
         end if;

         Get_Name_String (P.Node_Name);
         Add_Str_To_Name_Buffer ("_subprograms.adb");

         if Is_Regular_File (Name_Buffer (1 .. Name_Len)) then
            Write_Indentation;
            Write_Str ("for Body          (""Subprograms"") use """);
            Write_Name (P.Node_Name);
            Write_Line ("_subprograms.adb"";");
         end if;

         Get_Name_String (P.Appli_Name);
         Add_Str_To_Name_Buffer ("_marshallers.ads");

         if Is_Regular_File ("../common/" & Name_Buffer (1 .. Name_Len)) then
            Write_Indentation;
            Write_Str ("for Specification (""Marshallers"") use """);
            Write_Name (P.Appli_Name);
            Write_Line ("_marshallers.ads"";");
         end if;

         Write_Eol;
         Write_Indentation;
         Write_Line ("--  Custom middleware file names");
         Write_Eol;

         case P.Execution_Platform is
            when Platform_LEON | Platform_ERC_32 =>
               Write_Indentation;
               Write_Line ("for Body (""Output_Low_Level"")"
                           & " use ""output_low_level_leon.adb"";");

            when others =>
               Write_Indentation;
               Write_Line ("for Body (""Output_Low_Level"")"
                           & " use ""output_low_level_native.adb"";");
         end case;

         case P.Transport_Layer is
            when Transport_Ethernet =>
               Write_Indentation;
               Write_Line
                 ("for Body (""PolyORB_HI.Transport_Low_Level"")"
                  & " use ""polyorb_hi-transport_low_level_sockets.adb"";");

            when Transport_SpaceWire =>
               Write_Indentation;
               Write_Line
                 ("for Body (""PolyORB_HI.Transport_Low_Level"")"
                  & " use ""polyorb_hi-transport_low_level_mts.adb"";");

               if P.Is_Server then
                  Write_Indentation;
                  Write_Line
                    ("for Body (""PolyORB_HI.Transport_Low_Level.Extras"")"
                     & " use "
                     & """polyorb_hi-transport_low_level-extras_mts.adb"";");
               end if;

            when others =>
               Write_Indentation;
               Write_Line
                 ("for Body (""PolyORB_HI.Transport_Low_Level"")"
                  & " use ""polyorb_hi-transport_low_level_dummy.adb"";");
         end case;

         if Length (P.Spec_Names) > 0 then
            Write_Eol;
            Write_Indentation;
            Write_Line ("--  Custom user spec names");
            Write_Eol;

            for J in Name_Tables.First .. Name_Tables.Last (P.Spec_Names) loop
               Write_Indentation;
               Write_Str ("for Specification (""");
               Write_Name (P.Spec_Names.Table (J));
               Write_Str (""") use """);
               Write_Name (P.Custom_Spec_Names.Table (J));
               Write_Line (""";");
            end loop;
         end if;

         if Length (P.Body_Names) > 0 then
            Write_Eol;
            Write_Indentation;
            Write_Line ("--  Custom user body names");
            Write_Eol;

            for J in Name_Tables.First .. Name_Tables.Last (P.Body_Names) loop
               Write_Indentation;
               Write_Str ("for Body (""");
               Write_Name (P.Body_Names.Table (J));
               Write_Str (""") use """);
               Write_Name (P.Custom_Body_Names.Table (J));
               Write_Line (""";");
            end loop;
         end if;

         Write_Eol;
         Decrement_Indentation;
         Write_Indentation;
         Write_Line ("end Naming;");
         Decrement_Indentation;
         Write_Str ("end ");
         Write_Name (P.Node_Name);
         Write_Line (";");
      end PolyORB_HI_Ada_Ada_Project_File;

      -----------------------------------
      -- PolyORB_HI_C_Ada_Project_File --
      -----------------------------------

      procedure PolyORB_HI_C_Ada_Project_File (P : Ada_Project_File_Type) is
         pragma Unreferenced (P);
      begin
         null;
      end PolyORB_HI_C_Ada_Project_File;

      --------------------------------------
      -- PolyORB_Qos_Ada_Ada_Project_File --
      --------------------------------------

      procedure PolyORB_Qos_Ada_Ada_Project_File (P : Ada_Project_File_Type) is
      begin
         Write_Str ("project ");
         Write_Name (P.Node_Name);
         Write_Line (" is");
         Increment_Indentation;

         --  The source directory list

         Write_Indentation;
         Write_Line ("for Source_Dirs use");
         Increment_Indentation;

         Write_Indentation (-1);
         Write_Line ("(""."",");

         --  The PolyORB-QoS Ocarina runtime source directory is
         --  obtained using the `ocarina-config --runtime=polyorb`
         --  command output.

         declare
            Ocarina_Config : constant String := "ocarina-config";
            C              : String_Access := Locate_Exec_On_Path
              (Ocarina_Config);
            Args           : Argument_List :=
              (1 => new String'("--runtime=polyorb"));
            Status         : aliased Integer := 0;
            Path           : String_Access := Getenv ("PATH");
         begin
            if C = null then
               for I in Args'Range loop
                  Free (Args (I));
               end loop;

               Display_Error
                 ("Command not found: " & Ocarina_Config,
                  Fatal => True);
            else
               pragma Debug
                 (Display_Debug_Message (Ocarina_Config & ": " & C.all, True));
               Free (C);
            end if;

            pragma Debug (Display_Debug_Message ("PATH=" & Path.all, True));
            Free (Path);

            Write_Indentation;
            Write_Str
              (""""
               & GNAT.Expect.Get_Command_Output
               (Command    => Ocarina_Config,
                Arguments  => Args,
                Input      => "",
                Status     => Status'Access,
                Err_To_Out => True)
              & """");

            for I in Args'Range loop
               Free (Args (I));
            end loop;
         end;

         --  The user provided source dirs

         if Length (P.User_Source_Dirs) > 0 then
            Write_Line (",");

            for J in
              Name_Tables.First .. Name_Tables.Last (P.User_Source_Dirs) loop
               Write_Indentation;
               Write_Char ('"');
               Write_Name (P.User_Source_Dirs.Table (J));
               Write_Char ('"');

               exit when J = Name_Tables.Last (P.User_Source_Dirs);

               Write_Line (",");
            end loop;
         end if;

         Write_Line (");");

         Decrement_Indentation;

         --  The main subprogram name

         Write_Eol;
         Write_Indentation;
         Write_Str ("for Main use (""");
         Write_Name (P.Node_Name);
         Write_Line (".adb"");");

         --  The executables are put in the '../bin' directory

         Write_Indentation;
         Write_Line ("for Exec_Dir use ""../bin"";");

         --  Create the 'bin' directory

         Create_Directory (Get_String_Name ("../bin"));

         --  The compile flags

         Write_Eol;
         Write_Indentation;
         Write_Line ("Package Compiler is");
         Increment_Indentation;
         Write_Eol;

         Write_Indentation;
         Write_Line ("for Default_Switches (""Ada"") use (""-gnatwae"");");
         Write_Eol;

         Decrement_Indentation;
         Write_Indentation;
         Write_Line ("end Compiler;");

         --  The build flags

         Write_Eol;
         Write_Indentation;
         Write_Line ("Package Builder is");
         Increment_Indentation;
         Write_Eol;

         Write_Indentation;
         Write_Line ("for Default_Switches (""Ada"") use (""-m"");");
         Write_Eol;

         Decrement_Indentation;
         Write_Indentation;
         Write_Line ("end Builder;");

         Decrement_Indentation;
         Write_Str ("end ");
         Write_Name (P.Node_Name);
         Write_Line (";");
      end PolyORB_Qos_Ada_Ada_Project_File;
   end Ada_Project_Files;

   -----------
   -- Reset --
   -----------

   procedure Reset is
   begin
      Makefiles.Reset;
      Ada_Project_Files.Reset;
   end Reset;

end Ocarina.Generators.Build_Utils;
