------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--                  OCARINA.GENERATORS.PO_HI_ADA.ACTIVITY                   --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2006-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Namet;

with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Entities.Components;

with Ocarina.Generators.Utils;
with Ocarina.Generators.Properties;
with Ocarina.Generators.Messages;
with Ocarina.Generators.Ada_Tree.Nutils;
with Ocarina.Generators.Ada_Tree.Nodes;
with Ocarina.Generators.Ada_Values;
with Ocarina.Generators.PO_HI_Ada.Mapping;
with Ocarina.Generators.PO_HI_Ada.Runtime;

package body Ocarina.Generators.PO_HI_Ada.Activity is

   use Namet;
   use Ocarina.Nodes;
   use Ocarina.Entities.Components;
   use Ocarina.Generators.Utils;
   use Ocarina.Generators.Properties;
   use Ocarina.Generators.Messages;
   use Ocarina.Generators.Ada_Tree.Nutils;
   use Ocarina.Generators.Ada_Values;
   use Ocarina.Generators.PO_HI_Ada.Mapping;
   use Ocarina.Generators.PO_HI_Ada.Runtime;

   package AAU renames Ocarina.Nutils;
   package ADN renames Ocarina.Generators.Ada_Tree.Nodes;

   ------------------
   -- Package_Spec --
   ------------------

   package body Package_Spec is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);

      function Cyclic_Task_Instantiation (E : Node_Id) return Node_Id;
      --  Creates a package instantiation for a cyclic task

      function Cyclic_Task_Job_Spec (E : Node_Id) return Node_Id;
      --  Creates the parameterless subprogram specification that does
      --  the thread's job.

      -------------------------------
      -- Cyclic_Task_Instantiation --
      -------------------------------

      function Cyclic_Task_Instantiation (E : Node_Id) return Node_Id is
         N              : Node_Id;
         Parameter_List : constant List_Id :=
           New_List (ADN.K_Parameter_Profile);
         I              : Unsigned_Long_Long;
         S              : constant Node_Id := Parent_Subcomponent (E);
      begin
         --  Build the parameters list

         --  The entity name

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier
              (PN (P_Entity)),
            Actual_Parameter => Make_Defining_Identifier
              (Map_Ada_Enumerator_Name (S)));
         Append_Node_To_List (N, Parameter_List);

         --  The task period, the thread being periodic it has
         --  necessarily a period. Otherwise, an error will be raised.

         N := Map_Ada_Time (Get_Thread_Period (E));

         if No (N) then
            Display_Located_Error
              (Loc (E),
               "Unable to convert picoseconds period into nanoseconds",
               Fatal => True);
         end if;

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier (PN (P_Task_Period)),
            Actual_Parameter => N);
         Append_Node_To_List (N, Parameter_List);

         --  The task priority, if the thread has no priority, we
         --  assign a default one.

         I := Get_Thread_Priority (E);

         if I = 0 then
            N := RE (RE_Default_Priority);
         else
            N := Map_Ada_Priority (I);
         end if;

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier
              (PN (P_Task_Priority)),
            Actual_Parameter => N);
         Append_Node_To_List (N, Parameter_List);

         --  The task stack size, if the thread has no stack size, we
         --  assign a default one.

         I := Map_Ada_Size (Get_Thread_Stack_Size (E));

         if I = 0 then
            N := Make_Literal (New_Integer_Value (64_000, 1, 10));
         else
            N := Make_Literal (New_Integer_Value (I, 1, 10));
         end if;

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier
              (PN (P_Task_Stack_Size)),
            Actual_Parameter => N);
         Append_Node_To_List (N, Parameter_List);

         --  The task job

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier (PN (P_Job)),
            Actual_Parameter => Map_Task_Job_Identifier (S));
         Append_Node_To_List (N, Parameter_List);

         --  Build the package instantiation

         N := Make_Package_Instantiation
           (Defining_Identifier => Map_Cyclic_Task_Identifier (S),
            Generic_Package     => RU (RU_PolyORB_HI_Cyclic_Task),
            Parameter_List      => Parameter_List);
         return N;
      end Cyclic_Task_Instantiation;

      --------------------------
      -- Cyclic_Task_Job_Spec --
      --------------------------

      function Cyclic_Task_Job_Spec (E : Node_Id) return Node_Id is
         N : Node_Id;
         S : constant Node_Id := Parent_Subcomponent (E);
      begin
         N := Make_Subprogram_Specification
           (Defining_Identifier => Map_Task_Job_Identifier (S),
            Parameter_Profile   => No_List);
         return N;
      end Cyclic_Task_Job_Spec;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         U : constant Node_Id := ADN.Distributed_Application_Unit
           (ADN.Naming_Node (Backend_Node (Identifier (E))));
         P : constant Node_Id := ADN.Entity (U);
         S : Node_Id;
         N : Node_Id;
      begin
         Push_Entity (P);
         Push_Entity (U);
         Set_Activity_Spec;

         --  Make the deployment and naming packages use-visible

         Add_With_Package (RU (RU_Deployment, False), Used => True);

         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  If the process has a data subcomponent, then map a
               --  shared variable.

               if Utils.Is_Data (Corresponding_Instance (S)) then
                  N := Make_Object_Declaration
                    (Defining_Identifier => Map_Ada_Defining_Identifier (S),
                     Object_Definition   => Map_Ada_Data_Type_Designator
                       (Corresponding_Instance (S)));
                  Append_Node_To_List (N, ADN.Visible_Part (Current_Package));

                  --  Link the variable and the object

                  Bind_AADL_To_Object (Identifier (S), N);
               else

                  --  Visit the component instance corresponding to the
                  --  subcomponent S.

                  Visit (Corresponding_Instance (S));
               end if;

               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; -- U
         Pop_Entity; -- P
      end Visit_Process_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         Push_Entity (HI_Distributed_Application_Root);

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         P : constant Supported_Thread_Dispatch_Protocol :=
           Get_Thread_Dispatch_Protocol (E);
         S : constant Node_Id := Parent_Subcomponent (E);
         N : Node_Id;
      begin
         --  For each periodic thread, we define the spec of the Job
         --  subprogram. and instantiate a Cyclic_Task.

         if P = Thread_Periodic then
            N := Message_Comment
              ("Periodic task : "
               & Get_Name_String (Display_Name (Identifier (S))));
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));

            --  Create the spec of the parameterless subprogram that
            --  executes the thread job.

            N := Cyclic_Task_Job_Spec (E);
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));
            Bind_AADL_To_Job (Identifier (S), N);

            --  Instantiate the cyclic task

            N := Cyclic_Task_Instantiation (E);
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));
         end if;
      end Visit_Thread_Instance;

   end Package_Spec;

   ------------------
   -- Package_Body --
   ------------------

   package body Package_Body is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);

      function Cyclic_Task_Job_Body (E : Node_Id) return Node_Id;
      --  Creates the parameterless subprogram body that does the
      --  thread's job.

      Current_System_Instance : Node_Id := No_Node;
      --  Points to the current root system

      --------------------------
      -- Cyclic_Task_Job_Body --
      --------------------------

      function Cyclic_Task_Job_Body (E : Node_Id) return Node_Id is
         S              : constant Node_Id := Parent_Subcomponent (E);
         Spec           : constant Node_Id := ADN.Job_Node
           (Backend_Node (Identifier (S)));
         Declarations   : constant List_Id := New_List
           (ADN.K_Declaration_List);
         Statements     : constant List_Id := New_List (ADN.K_Statement_List);
         C_Statements   : List_Id;
         Src_List       : List_Id;
         Src_Thread     : Node_Id;
         N              : Node_Id;
         P              : Node_Id;
         Call_Seq       : Node_Id;
         Spg_Call       : Node_Id;
         Req_Name       : Name_Id := PN (P_Req);
         Req_Expression : Node_Id;
      begin
         if Has_In_Ports (E) then
            --  If the periodic thread has IN ports, we check whether
            --  it received a message in these ports.

            --  The reception of the message is performed using non
            --  clocking routine. A boolean flog will indicate whether
            --  a reception really occured or not.

            --  Declare the Message variable

            N := Make_Object_Declaration
              (Defining_Identifier => Make_Defining_Identifier
                 (PN (P_Incoming_Message)),
               Object_Definition   => RE (RE_Message_Type),
               Aliased_Present     => True);
            Append_Node_To_List (N, Declarations);

            --  Declare the boolean flag

            N := Make_Object_Declaration
              (Defining_Identifier => Make_Defining_Identifier
                 (PN (P_Got_Data)),
               Object_Definition   => RE (RE_Boolean));
            Append_Node_To_List (N, Declarations);

            --  Get the sender thread

            P := First_Node (Features (E));

            while Present (P) loop
               if Kind (P) = K_Port_Spec_Instance and then Is_In (P) then
                  --  FIXME: For now, we support only periodic threads
                  --  that are the destination of exactly one thread.

                  Src_List := Get_Source_Ports (P);

                  if AAU.Length (Src_List) /= 1 then
                     Display_Located_Error
                       (Loc (P),
                        "IN event ports that are destinations of multiple"
                        & " connections not supported yet.",
                        Fatal => True);
                  end if;

                  --  Now we have all the enumerators of type
                  --  Thread_Type that are sources of the port P

                  --  FIXME: When multiple source connections are
                  --  supported, use multiple sub-stubs to handle each
                  --  incompin submessage.

                  --  Fetch the thread corresponding to the source
                  --  port.

                  Src_Thread := Parent_Component
                    (Item (First_Node (Src_List)));

                  exit; --  FIXME: One destination thread
               end if;

               P := Next_Node (P);
            end loop;

            --  Call the non blocking routine

            N := Make_Subprogram_Call
              (RE (RE_Non_Blocking_Receive),
               Make_List_Id
               (Make_Defining_Identifier
                (Map_Ada_Enumerator_Name
                 (Parent_Subcomponent
                  (Src_Thread))),
                Make_Defining_Identifier (PN (P_Incoming_Message)),
                Make_Defining_Identifier (PN (P_Got_Data))));
            Append_Node_To_List (N, Statements);

            --  All the subprogram calls this thread is supposed to
            --  perform will be done only if it received a message in
            --  its ports.

            --  FIXME: This design has to be refined as follows: if
            --  the thread does not receive a message, it can call the
            --  subprogram that are not affected by the absence of
            --  these thread.

            C_Statements := New_List (ADN.K_Statement_List);

            --  The first statement in the IF cloc, will be the
            --  unmarshalling of the request from the message.

            --  Declare the request '<Thread>_Job_Req'. We declare the
            --  variable directly in the paquage statements. This is
            --  not harmful because the only thread that accesses this
            --  variable is <Thread>. This variable is initialized
            --  with a default value to avoid livelocking between
            --  threads that wait for each others.

            --  Build the request default value

            --  Get the subprogram wrapper of the thread

            if not AAU.Is_Empty (Calls (E)) then
               if AAU.Is_Empty (Subprogram_Calls (First_Node (Calls (E)))) then
                  Display_Located_Error
                    (Loc (First_Node (Calls (E))),
                     "Empty call sequence",
                     Fatal   => False,
                     Warning => True);
               end if;

               declare
                  Wrapper_Spg : constant Node_Id :=
                    Corresponding_Instance
                    (First_Node
                     (Subprogram_Calls
                      (First_Node
                       (Calls
                        (E)))));

                  Aggregates  : constant List_Id :=
                    New_List (ADN.K_Component_List);
                  F           : Node_Id;
               begin
                  --  Add the record discriminant

                  N := Make_Component_Association
                    (Make_Defining_Identifier (PN (P_Operation)),
                     Make_Defining_Identifier
                     (Map_Ada_Enumerator_Name
                      (Wrapper_Spg)));
                  Append_Node_To_List (N, Aggregates);

                  --  For each IN parameter of the subprogram, create
                  --  a component association.

                  if not AAU.Is_Empty (Features (Wrapper_Spg)) then
                     F := First_Node (Features (Wrapper_Spg));

                     while Present (F) loop
                        if Kind (F) = K_Parameter_Instance then
                           if Is_In (F) and then not Is_Out (F) then
                              N := Make_Component_Association
                                (Make_Defining_Identifier
                                 (Map_Ada_Full_Parameter_Name
                                  (Wrapper_Spg, F, 'C')),
                                 Get_Ada_Default_Value
                                 (Corresponding_Instance (F)));
                              Append_Node_To_List (N, Aggregates);
                           end if;
                        end if;

                        F := Next_Node (F);
                     end loop;
                  end if;

                  N := Extract_Designator
                    (ADN.Request_Type_Node
                     (Backend_Node
                      (Identifier
                       (Current_System_Instance))));

                  Req_Expression := Make_Qualified_Expression
                    (Subtype_Mark => N,
                     Aggregate    => Make_Record_Aggregate (Aggregates));
               end;
            else
               Req_Expression := No_Node;
            end if;

            --  For more code readability, add a use clause to the
            --  <App>_Internals package

            N := Extract_Designator
              (ADN.Internals_Node
               (Backend_Node
                (Identifier
                 (Current_System_Instance))));
            Add_With_Package (N, Used => True);

            --  Build the variable name

            Get_Name_String (ADN.Name (Map_Task_Job_Identifier (S)));
            Add_Str_To_Name_Buffer ("_Req");
            Req_Name := Name_Find;

            N := Make_Object_Declaration
              (Defining_Identifier => Make_Defining_Identifier (Req_Name),
               Object_Definition   => Extract_Designator
                 (ADN.Request_Type_Node
                  (Backend_Node
                   (Identifier
                    (Current_System_Instance)))),
               Expression          => Req_Expression);
            Append_Node_To_List (N, ADN.Statements (Current_Package));

            --  Unmarshall the request

            N := Make_Assignment_Statement
              (Make_Defining_Identifier (Req_Name),
               Make_Subprogram_Call
               (RE (RE_Unmarshall_Request),
                Make_List_Id (Make_Designator (PN (P_Incoming_Message)))));
            Append_Node_To_List (N, C_Statements);

            --  Build the global IF statement

            N := Make_If_Statement
              (Condition       => Make_Defining_Identifier (PN (P_Got_Data)),
               Then_Statements => C_Statements);
            Append_Node_To_List (N, Statements);
         else
            --  The thread has no IN ports, there is no need to
            --  generate an IF statement.

            C_Statements := Statements;
         end if;

         --  Visit all the call sequences of the thread

         if not AAU.Is_Empty (Calls (E)) then
            Check_Thread_Consistency (E);

            Call_Seq := First_Node (Calls (E));
            Handle_Call_Sequence
              (E, Call_Seq, Declarations, Statements, Req_Name);

            --  Trigger all remote threads

            if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then

               Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

               while Present (Spg_Call) loop
                  Call_Remote_Subprograms
                    (E, Spg_Call, Declarations, Statements);

                  Spg_Call := Next_Node (Spg_Call);
               end loop;
            end if;
         end if;

         N := Make_Subprogram_Implementation
           (Spec, Declarations, Statements);
         return N;
      end Cyclic_Task_Job_Body;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         U : constant Node_Id := ADN.Distributed_Application_Unit
           (ADN.Naming_Node (Backend_Node (Identifier (E))));
         P : constant Node_Id := ADN.Entity (U);
         S : Node_Id;
      begin
         Push_Entity (P);
         Push_Entity (U);
         Set_Activity_Body;

         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; -- U
         Pop_Entity; -- P
      end Visit_Process_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         Push_Entity (HI_Distributed_Application_Root);

         Current_System_Instance := E;

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         P : constant Supported_Thread_Dispatch_Protocol :=
           Get_Thread_Dispatch_Protocol (E);
         S : constant Node_Id := Parent_Subcomponent (E);
         N : Node_Id;
      begin
         --  For each periodic thread, we define the spec of the Job
         --  subprogram. and instantiate a Cyclic_Task.

         if P = Thread_Periodic then
            N := Message_Comment
              ("Periodic task : "
               & Get_Name_String (Display_Name (Identifier (S))));
            Append_Node_To_List (N, ADN.Statements (Current_Package));

            --  Create the body of the parameterless subprogram that
            --  executes the thread job.

            N := Cyclic_Task_Job_Body (E);
            Append_Node_To_List (N, ADN.Statements (Current_Package));
            Bind_AADL_To_Job (Identifier (S), N);
         end if;
      end Visit_Thread_Instance;

   end Package_Body;

end Ocarina.Generators.PO_HI_Ada.Activity;
