------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--    O C A R I N A . G E N E R A T O R S . P O _ H I _ A D A . C S R A     --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2006-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Entities.Components;

with Ocarina.Generators.Utils;
with Ocarina.Generators.Ada_Tree.Nutils;
with Ocarina.Generators.Ada_Tree.Nodes;
with Ocarina.Generators.PO_HI_Ada.Runtime;
with Ocarina.Generators.PO_HI_Ada.Mapping;

package body Ocarina.Generators.PO_HI_Ada.CSRA is

   use Ocarina.Nodes;
   use Ocarina.Entities.Components;
   use Ocarina.Generators.Utils;
   use Ocarina.Generators.Ada_Tree.Nutils;
   use Ocarina.Generators.PO_HI_Ada.Runtime;
   use Ocarina.Generators.PO_HI_Ada.Mapping;

   package AAU renames Ocarina.Nutils;
   package ADN renames Ocarina.Generators.Ada_Tree.Nodes;

   ------------------
   -- Package_Spec --
   ------------------

   package body Package_Spec is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Subprogram_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);

      function Stub_Spec (E : Node_Id) return Node_Id;
      --  Generates the spec of the stub subprogram corresponding to
      --  the subprogram instance E.

      ---------------
      -- Stub_Spec --
      ---------------

      function Stub_Spec (E : Node_Id) return Node_Id is
         N       : Node_Id;
         F       : Node_Id;
         Profile : constant List_Id := New_List (ADN.K_Parameter_Profile);
      begin
         --  FIXME : All this has to be refined in order to support
         --  multi port threads that are connected to several source
         --  threads.

         --  The 'From' parameter

         N := Make_Parameter_Specification
           (Defining_Identifier => Make_Defining_Identifier (PN (P_From)),
            Subtype_Mark        => RE (RE_Entity_Type));
         Append_Node_To_List (N, Profile);

         --  The 'To' parameter

         N := Make_Parameter_Specification
           (Defining_Identifier => Make_Defining_Identifier (PN (P_To)),
            Subtype_Mark        => RE (RE_Entity_Type));
         Append_Node_To_List (N, Profile);

         --  The 'Message' parameter

         N := Make_Parameter_Specification
           (Defining_Identifier => Make_Defining_Identifier (PN (P_Message)),
            Subtype_Mark        => RE (RE_Message_Type),
            Parameter_Mode      => Mode_Inout);
         Append_Node_To_List (N, Profile);

         --  The subprogram call actual parameters

         if not AAU.Is_Empty (Features (E)) then
            F := First_Node (Features (E));

            while Present (F) loop
               if Kind (F) = K_Parameter_Instance then
                  if Is_In (F) and then not Is_Out (F) then
                     N := Make_Parameter_Specification
                       (Defining_Identifier => Map_Ada_Defining_Identifier (F),
                        Subtype_Mark        => Map_Ada_Data_Type_Designator
                          (Corresponding_Instance (F)),
                        Parameter_Mode      => Mode_In);
                     Append_Node_To_List (N, Profile);
                  end if;

               end if;

               F := Next_Node (F);
            end loop;
         end if;

         --  Build the spec

         N := Make_Subprogram_Specification
           (Defining_Identifier => Map_Stub_Identifier (E),
            Parameter_Profile   => Profile);
         return N;
      end Stub_Spec;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when CC_Subprogram =>
               Visit_Subprogram_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));

            while Present (S) loop
               Visit (Corresponding_Instance (S));

               S := Next_Node (S);
            end loop;
         end if;
      end Visit_Process_Instance;

      -------------------------------
      -- Visit_Subprogram_Instance --
      -------------------------------

      procedure Visit_Subprogram_Instance (E : Node_Id) is
         N : Node_Id;
      begin
         --  We create a stub only for subprograms that can be called
         --  remotly in an asynchronous way, namely, subprograms that
         --  have IN parameters and *do not* have out parameters.

         --  Make sure that each subprogram is handled only once

         if No (Get_Handling (E, By_Name, H_Ada_Stub_Spec)) then
            if Has_In_Parameters (E) then
               N := Stub_Spec (E);
               Set_Handling (E, By_Name, H_Ada_Stub_Spec, N);
               Append_Node_To_List (N, ADN.Visible_Part (Current_Package));
            end if;
         end if;

         if Has_In_Parameters (E) then
            Bind_AADL_To_Stub
              (Identifier (E),
               Get_Handling (E, By_Name, H_Ada_Stub_Spec));
         end if;
      end Visit_Subprogram_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
         U : constant Node_Id := ADN.Distributed_Application_Unit
           (ADN.Internals_Node (Backend_Node (Identifier (E))));
      begin
         Push_Entity (HI_Distributed_Application_Root);
         Push_Entity (U);
         Set_CSRA_Spec;

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; --  U
         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         Call_Seq : Node_Id;
         Spg_Call : Node_Id;
      begin
         --  Visit all the call sequences of the thread

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;
      end Visit_Thread_Instance;

   end Package_Spec;

   ------------------
   -- Package_Body --
   ------------------

   package body Package_Body is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Subprogram_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);

      function Stub_Body (E : Node_Id) return Node_Id;
      --  Generates the body of the stub subprogram corresponding to
      --  the subprogram instance E.

      ---------------
      -- Stub_Body --
      ---------------

      function Stub_Body (E : Node_Id) return Node_Id is
         Spec         : constant Node_Id := ADN.Stub_Node
           (Backend_Node (Identifier (E)));
         Declarations : constant List_Id := New_List (ADN.K_Declaration_List);
         Statements   : constant List_Id := New_List (ADN.K_Statement_List);
         Components   : constant List_Id := New_List (ADN.K_Component_List);
         N            : Node_Id;
         F            : Node_Id;
      begin
         --  Marshall

         N := Message_Comment ("Marshall the request");
         Append_Node_To_List (N, Statements);

         --  Build the record aggregate

         N := Make_Component_Association
           (Selector_Name => Make_Defining_Identifier (PN (P_Operation)),
            Expression    => Copy_Node
              (ADN.Enumerator_Node
               (Backend_Node
                (Identifier (E)))));
         Append_Node_To_List (N, Components);

         --  We use the parameters of the corresponding subprogram to
         --  generate the correct full selector name.

         if not AAU.Is_Empty (Features (E)) then
            F := First_Node (Features (E));

            while Present (F) loop
               if Kind (F) = K_Parameter_Instance then
                  if Is_In (F) and then not Is_Out (F) then
                     N := Make_Component_Association
                       (Selector_Name => Make_Defining_Identifier
                          (Map_Ada_Full_Parameter_Name (E, F, 'C')),
                        Expression    => Map_Ada_Defining_Identifier (F));
                     Append_Node_To_List (N, Components);
                  end if;
               end if;

               F := Next_Node (F);
            end loop;
         end if;

         N := Make_Subprogram_Call
           (RE (RE_Marshall_Request),
            Make_List_Id (Make_Record_Aggregate (Components),
                          Make_Designator (PN (P_Message))));
         Append_Node_To_List (N, Statements);

         --  Send

         N := Message_Comment ("Send the message");
         Append_Node_To_List (N, Statements);

         N := Make_Subprogram_Call
           (RE (RE_Send),
            Make_List_Id (Make_Designator (PN (P_From)),
                          Make_Designator (PN (P_To)),
                          Make_Designator (PN (P_Message))));
         Append_Node_To_List (N, Statements);

         --  Reallocate message

         N := Message_Comment ("Reallocate the message");
         Append_Node_To_List (N, Statements);

         N := Make_Subprogram_Call
           (RE (RE_Reallocate),
            Make_List_Id (Make_Designator (PN (P_Message))));
         Append_Node_To_List (N, Statements);

         --  Make the subprogram body

         N := Make_Subprogram_Implementation
           (Spec, Declarations, Statements);
         return N;
      end Stub_Body;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when CC_Subprogram =>
               Visit_Subprogram_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));

            while Present (S) loop
               Visit (Corresponding_Instance (S));

               S := Next_Node (S);
            end loop;
         end if;
      end Visit_Process_Instance;

      -------------------------------
      -- Visit_Subprogram_Instance --
      -------------------------------

      procedure Visit_Subprogram_Instance (E : Node_Id) is
         N : Node_Id;
      begin
         --  We create a stub only for subprograms that can be called
         --  remotly in an asynchronous way, namely, subprograms that
         --  have IN parameters and *do not* have out parameters.

         --  Make sure that each subprogram is handled only once

         if No (Get_Handling (E, By_Name, H_Ada_Stub_Body)) then

            if Has_In_Parameters (E) then
               N := Stub_Body (E);
               Set_Handling (E, By_Name, H_Ada_Stub_Body, N);
               Append_Node_To_List (N, ADN.Statements (Current_Package));
            end if;
         end if;
      end Visit_Subprogram_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
         U : constant Node_Id := ADN.Distributed_Application_Unit
           (ADN.Internals_Node (Backend_Node (Identifier (E))));
      begin
         Push_Entity (HI_Distributed_Application_Root);
         Push_Entity (U);
         Set_CSRA_Body;

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         --  For more readability, we add a use clause to the _Internals
         --  package at the beginning of the package statements

         if not Is_Empty (ADN.Statements (Current_Package)) then
            Add_With_Package
              (Extract_Designator
               (ADN.Internals_Package
                (Current_Entity)),
              Used => True);
         end if;

         Pop_Entity; --  U
         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         Call_Seq : Node_Id;
         Spg_Call : Node_Id;
      begin
         --  Visit all the call sequences of the thread

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;
      end Visit_Thread_Instance;

   end Package_Body;

end Ocarina.Generators.PO_HI_Ada.CSRA;
