------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--                  OCARINA.GENERATORS.PO_HI_ADA.HANDLERS                   --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2006-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Namet;

with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Entities.Components;

with Ocarina.Generators.Utils;
with Ocarina.Generators.Messages;
with Ocarina.Generators.Properties;
with Ocarina.Generators.Ada_Tree.Nutils;
with Ocarina.Generators.Ada_Tree.Nodes;
with Ocarina.Generators.PO_HI_Ada.Mapping;
with Ocarina.Generators.PO_HI_Ada.Runtime;
with Ocarina.Generators.Ada_Values;

package body Ocarina.Generators.PO_HI_Ada.Handlers is

   use Namet;
   use Ocarina.Nodes;
   use Ocarina.Entities.Components;
   use Ocarina.Generators.Utils;
   use Ocarina.Generators.Messages;
   use Ocarina.Generators.Properties;
   use Ocarina.Generators.Ada_Tree.Nutils;
   use Ocarina.Generators.PO_HI_Ada.Mapping;
   use Ocarina.Generators.PO_HI_Ada.Runtime;
   use Ocarina.Generators.Ada_Values;

   package AAU renames Ocarina.Nutils;
   package ADN renames Ocarina.Generators.Ada_Tree.Nodes;

   ------------------
   -- Package_Spec --
   ------------------

   package body Package_Spec is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);

      Current_DA : Node_Id;
      --  The current AADL distributed application node

      function Sporadic_Task_Instantiation (E : Node_Id) return Node_Id;
      --  Build a package instantiation of a handler task
      --  corresponding to the aperiodic thread E.

      ---------------------------------
      -- Sporadic_Task_Instantiation --
      ---------------------------------

      function Sporadic_Task_Instantiation (E : Node_Id) return Node_Id is
         N                  : Node_Id;
         Associations       : constant List_Id := New_List (ADN.K_List_Id);
         P                  : Node_Id;
         Has_In_Event_Ports : Boolean := False;
         Src_List           : List_Id;
         I                  : Unsigned_Long_Long;
         Src_Thread         : Node_Id;
         S                  : constant Node_Id := Parent_Subcomponent (E);
      begin
         --  Handle the type discriminants

         --  * Entity

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier
              (PN (P_Entity)),
            Actual_Parameter => Make_Defining_Identifier
              (Map_Ada_Enumerator_Name (Parent_Subcomponent (E))));
         Append_Node_To_List (N, Associations);

         --  * From

         --  Raise an error if the thread does not have IN ports

         if not Has_In_Ports (E) then
            Display_Located_Error
              (Loc (E),
               "This sporadic thread does not have IN ports",
               Fatal => True);
         end if;

         P := First_Node (Features (E));

         while Present (P) loop
            if Kind (P) = K_Port_Spec_Instance
              and then Is_In (P)
              and then Is_Event (P)
            then
               Has_In_Event_Ports := True;

               --  FIXME: For now, we support only sporadic threads
               --  that are triggered by exactly one in event [data]
               --  port and their unique in event [data] port is the
               --  destination of a unique connection.

               Src_List := Get_Source_Ports (P);

               if AAU.Length (Src_List) /= 1 then
                  Display_Located_Error
                    (Loc (P),
                     "IN event ports that are destinations of multiple"
                     & " connections not supported yet.",
                     Fatal => True);
               end if;

               --  Now we have all the enumerators of type Thread_Type
               --  that are sources of the port P

               --  FIXME: When multiple source connections are
               --  supported, the expression of the association should
               --  be a static array of all the sources

               --  Fetch the thread corresponding to the source port

               Src_Thread := Parent_Component (Item (First_Node (Src_List)));

               N := Make_Parameter_Association
                 (Selector_Name    => Make_Defining_Identifier (CN (C_From)),
                  Actual_Parameter => Make_Defining_Identifier
                    (Map_Ada_Enumerator_Name
                     (Parent_Subcomponent
                      (Src_Thread))));
               Append_Node_To_List (N, Associations);

               exit; --  FIXME: One unique in event [data] port
            end if;

            P := Next_Node (P);
         end loop;

         --  Display an error if the thread does not have 'in event'
         --  ports.

         if not Has_In_Event_Ports then
            Display_Located_Error
              (Loc (E),
               "No one of the IN ports of this sporadic is an event port",
               Fatal => True);
         end if;

         --  FIXME: For now, we support only sporadic threads that are
         --  triggered by exactly one event [data] port.

         if not Has_In_Ports (E) then
            Display_Located_Error
              (Loc (E),
               "This sporadic thread does not have IN ports",
               Fatal => True);
         end if;

         --  * Task_Period

         --  The task inter-arrival time, the thread being sporadic it
         --  has necessarily an minimal interarrival time. Otherwise,
         --  an error will be raised.

         N := Map_Ada_Time (Get_Thread_Period (E));

         if No (N) then
            Display_Located_Error
              (Loc (E),
               "Unable to convert picoseconds period into nanoseconds",
               Fatal => True);
         end if;

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier (PN (P_Task_Period)),
            Actual_Parameter => N);
         Append_Node_To_List (N, Associations);

         --  * Task_Priority

         --  The task priority, if the thread has no priority, we
         --  assign a default one.

         I := Get_Thread_Priority (E);

         if I = 0 then
            N := RE (RE_Default_Priority);
         else
            N := Map_Ada_Priority (I);
         end if;

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier
              (PN (P_Task_Priority)),
            Actual_Parameter => N);
         Append_Node_To_List (N, Associations);

         --  * Task_Stack_Size

         --  The task stack size, if the thread has no stack size, we
         --  assign a default one.

         I := Map_Ada_Size (Get_Thread_Stack_Size (E));

         if I = 0 then
            N := Make_Literal (New_Integer_Value (64_000, 1, 10));
         else
            N := Make_Literal (New_Integer_Value (I, 1, 10));
         end if;

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier
              (PN (P_Task_Stack_Size)),
            Actual_Parameter => N);
         Append_Node_To_List (N, Associations);

         --  * Dispatcher

         N := Extract_Designator
           (ADN.Process_Request_Node
            (Backend_Node
             (Identifier
              (Current_DA))));

         N := Make_Parameter_Association
           (Selector_Name    => Make_Defining_Identifier (PN (P_Dispatcher)),
            Actual_Parameter => N);
         Append_Node_To_List (N, Associations);

         --  Build the package instantiation

         N := Make_Package_Instantiation
           (Defining_Identifier => Map_Handler_Identifier (S),
            Generic_Package     => RU (RU_PolyORB_HI_Sporadic_Task),
            Parameter_List      => Associations);
         return N;
      end Sporadic_Task_Instantiation;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         U : constant Node_Id := ADN.Distributed_Application_Unit
           (ADN.Naming_Node (Backend_Node (Identifier (E))));
         P : constant Node_Id := ADN.Entity (U);
         S : Node_Id;
      begin
         Push_Entity (P);
         Push_Entity (U);
         Set_Handlers_Spec;

         --  Make the deployment and naming packages use-visible

         Add_With_Package (RU (RU_Deployment, False), Used => True);

         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; -- U
         Pop_Entity; -- P
      end Visit_Process_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         Push_Entity (HI_Distributed_Application_Root);

         --  Set the global Current_DA variable

         Current_DA := E;

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         P : constant Supported_Thread_Dispatch_Protocol :=
           Get_Thread_Dispatch_Protocol (E);
         S : constant Node_Id := Parent_Subcomponent (E);
         N : Node_Id;
      begin
         --  For each sporadic thread, we declare a handler task

         if P = Thread_Sporadic then
            N := Message_Comment
              ("Sporadic task : "
               & Get_Name_String (Display_Name (Identifier (S))));
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));

            --  Declare the handler task

            N := Sporadic_Task_Instantiation (E);
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));
         end if;
      end Visit_Thread_Instance;

   end Package_Spec;

end Ocarina.Generators.PO_HI_Ada.Handlers;
