------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--                 OCARINA.GENERATORS.PO_HI_ADA.INTERNALS                   --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2006-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Entities.Components;

with Ocarina.Generators.Utils;
with Ocarina.Generators.PO_HI_Ada.Mapping;
with Ocarina.Generators.Ada_Tree.Nutils;
with Ocarina.Generators.Ada_Tree.Nodes;

package body Ocarina.Generators.PO_HI_Ada.Internals is

   use Ocarina.Nodes;
   use Ocarina.Entities.Components;
   use Ocarina.Generators.Utils;
   use Ocarina.Generators.PO_HI_Ada.Mapping;
   use Ocarina.Generators.Ada_Tree.Nutils;

   package AAU renames Ocarina.Nutils;
   package ADN renames Ocarina.Generators.Ada_Tree.Nodes;

   ------------------
   -- Package_Spec --
   ------------------

   package body Package_Spec is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Subprogram_Instance (E : Node_Id);

      --  Global variables for the generated entities. Note that it is
      --  safe to use global variable in this case because there is
      --  only one distributed application node and it is visited only
      --  once in this package.

      Enumerators : List_Id;
      Variants    : List_Id;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when CC_Subprogram =>
               Visit_Subprogram_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;
      end Visit_Process_Instance;

      -------------------------------
      -- Visit_Subprogram_Instance --
      -------------------------------

      procedure Visit_Subprogram_Instance (E : Node_Id) is
         Variant     : Node_Id;
         Choices     : List_Id;
         Choice      : Node_Id;
         Components  : List_Id;
         F           : Node_Id;
         N           : Node_Id;
         Call_Seq    : Node_Id;
         Spg_Call    : Node_Id;
      begin
         --  Visit all the call sequences of the subprogram

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;

         --  If the subprogram is remote-callable (if it has at least
         --  one IN parameter), then we add an enumerator
         --  corresponding to the subprogram to the enumerator list.

         if Has_In_Parameters (E) then
            --  Check whether we already handled the subprogram

            if No
              (Get_Handling
               (E,
                By_Name,
                H_Add_Enumerator_Internals_Spec))
            then
               --  Create an enumerator literal and append it to the
               --  distributed application enumerator list.

               N := Make_Defining_Identifier (Map_Ada_Enumerator_Name (E));
               Set_Handling (E, By_Name, H_Add_Enumerator_Internals_Spec, N);
               Append_Node_To_List (N, Enumerators);

               --  Bind the AADL and the Ada trees
            end if;

            --  Bind the AADL and the Ada trees

            Bind_AADL_To_Enumerator
              (Identifier (E),
               Get_Handling (E, By_Name, H_Add_Enumerator_Internals_Spec));

            if No
              (Get_Handling
               (E,
                By_Name,
                H_Add_Case_Alternative_Internals_Spec))
            then
               --  Create a swtich case variant and append it to the
               --  variant list of the distributed application.

               Variant := New_Node (ADN.K_Variant);
               Set_Handling
                 (E,
                  By_Name,
                  H_Add_Case_Alternative_Internals_Spec,
                  Variant);
               Append_Node_To_List (Variant, Package_Spec.Variants);
               Choices := New_List (ADN.K_Discrete_Choice_List);
               ADN.Set_Discrete_Choices (Variant, Choices);
               Components := New_List (ADN.K_Component_List);
               ADN.Set_Component_List (Variant, Components);

               Choice := Make_Defining_Identifier
                 (Map_Ada_Enumerator_Name (E));
               Append_Node_To_List (Choice, Choices);

               --  For each IN parameter of the subprogram, create a
               --  component declaration.

               if not AAU.Is_Empty (Features (E)) then
                  F := First_Node (Features (E));

                  while Present (F) loop
                     if Kind (F) = K_Parameter_Instance then
                        if Is_In (F) and then not Is_Out (F) then
                           N := Make_Component_Declaration
                             (Defining_Identifier => Make_Defining_Identifier
                                (Map_Ada_Full_Parameter_Name (E, F, 'C')),
                              Subtype_Indication  =>
                                Map_Ada_Data_Type_Designator
                                (Corresponding_Instance (F)));
                           Append_Node_To_List (N, Components);
                        end if;
                     end if;

                     F := Next_Node (F);
                  end loop;
               end if;
            end if;
         end if;
      end Visit_Subprogram_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
         U : Node_Id;
         N : Node_Id;
      begin
         Push_Entity (HI_Distributed_Application_Root);

         --  It is important to push the distributed application node
         --  to the top of the entity stack because it is used when
         --  mapping the HI_Units.

         U := Map_HI_Unit (E);
         Push_Entity (U);
         Set_Internals_Spec;

         --  Create the global lists

         Enumerators := New_List (ADN.K_Enumeration_Literals);
         Variants    := New_List (ADN.K_Variant_List);

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         --  If the application is distributes, generate the request
         --  discriminated record type.

         if not Is_Empty (Enumerators) then

            --  Create the enumeration type for all the operations of
            --  the distributed application.

            N := Message_Comment ("Enumeration type for all the operations"
                                  & " in the distributed application.");
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));

            N := Make_Full_Type_Declaration
              (Defining_Identifier => Make_Defining_Identifier
                 (TN (T_Operations)),
               Type_Definition     => Make_Enumeration_Type_Definition
                 (Enumerators));
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));

            --  Create the discriminated record which associates to
            --  each operation, its parameters.

            N := Message_Comment ("Record type for all the operation"
                                  & " parameters in the distributed"
                                  & " application.");
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));

            N := Make_Variant_Part
              (Discriminant => Make_Defining_Identifier (CN (C_Operation)),
               Variant_List => Variants);
            N := Make_Full_Type_Declaration
              (Defining_Identifier => Make_Defining_Identifier
                 (TN (T_Request)),
               Discriminant_Spec   => Make_Component_Declaration
                 (Defining_Identifier => Make_Defining_Identifier
                    (CN (C_Operation)),
                  Subtype_Indication  => Make_Defining_Identifier
                    (TN (T_Operations)),
                  Expression          => Make_Attribute_Designator
                    (Make_Designator (TN (T_Operations)),
                     A_First)),
               Type_Definition     => Make_Record_Type_Definition
                 (Make_Record_Definition (Make_List_Id (N))));
            Bind_AADL_To_Request_Type (Identifier (E), N);
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));
         end if;

         Pop_Entity; --  U
         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         Call_Seq : Node_Id;
         Spg_Call : Node_Id;
      begin
         --  Visit all the call sequences of the thread

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;
      end Visit_Thread_Instance;

   end Package_Spec;

   ------------------
   -- Package_Body --
   ------------------

   package body Package_Body is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         S : Node_Id;
      begin

         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

      end Visit_Process_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
         U : constant Node_Id := ADN.Distributed_Application_Unit
           (ADN.Internals_Node (Backend_Node (Identifier (E))));
      begin
         Push_Entity (HI_Distributed_Application_Root);
         Push_Entity (U);
         Set_Internals_Body;

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; --  U
         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         Call_Seq : Node_Id;
         Spg_Call : Node_Id;
      begin
         --  Visit all the call sequences of the thread

         if not AAU.Is_Empty (Calls (E)) then
            Call_Seq := First_Node (Calls (E));

            while Present (Call_Seq) loop
               --  For each call sequence visit all the called
               --  subprograms.

               if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                  Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                  while Present (Spg_Call) loop
                     Visit (Corresponding_Instance (Spg_Call));

                     Spg_Call := Next_Node (Spg_Call);
                  end loop;
               end if;

               Call_Seq := Next_Node (Call_Seq);
            end loop;
         end if;
      end Visit_Thread_Instance;

   end Package_Body;

end Ocarina.Generators.PO_HI_Ada.Internals;
