------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--    O C A R I N A . G E N E R A T O R S . P O _ H I _ A D A . S S R A     --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2006-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Entities.Components;

with Ocarina.Generators.Utils;
with Ocarina.Generators.Properties;
with Ocarina.Generators.Ada_Tree.Nutils;
with Ocarina.Generators.Ada_Tree.Nodes;
with Ocarina.Generators.PO_HI_Ada.Mapping;
with Ocarina.Generators.PO_HI_Ada.Runtime;

package body Ocarina.Generators.PO_HI_Ada.SSRA is

   use Ocarina.Nodes;
   use Ocarina.Entities.Components;
   use Ocarina.Generators.Utils;
   use Ocarina.Generators.Properties;
   use Ocarina.Generators.Ada_Tree.Nutils;
   use Ocarina.Generators.PO_HI_Ada.Mapping;
   use Ocarina.Generators.PO_HI_Ada.Runtime;

   package AAU renames Ocarina.Nutils;
   package ADN renames Ocarina.Generators.Ada_Tree.Nodes;

   ------------------
   -- Package_Spec --
   ------------------

   package body Package_Spec is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);

      function Process_Request_Spec return Node_Id;
      --  Creates the spec for the process request operation

      --------------------------
      -- Process_Request_Spec --
      --------------------------

      function Process_Request_Spec return Node_Id is
         N       : Node_Id;
         Profile : constant List_Id := New_List (ADN.K_Parameter_Profile);
      begin
         N := Make_Parameter_Specification
           (Defining_Identifier => Make_Defining_Identifier (PN (P_Message)),
            Subtype_Mark        => RE (RE_Message_Type),
            Parameter_Mode      => Mode_Inout);
         Append_Node_To_List (N, Profile);

         N := Make_Subprogram_Specification
           (Make_Defining_Identifier (SN (S_Process_Request)), Profile);
         return N;
      end Process_Request_Spec;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         N : Node_Id;
         U : constant Node_Id := ADN.Distributed_Application_Unit
           (ADN.Internals_Node (Backend_Node (Identifier (E))));
      begin
         Push_Entity (HI_Distributed_Application_Root);
         Push_Entity (U);
         Set_SSRA_Spec;

         --  If there is no request type generated for the
         --  application, we generate nothing

         if Present
           (ADN.Request_Type_Node
            (Backend_Node
             (Identifier
              (E))))
         then
            N := Make_Pragma_Statement (Pragma_Elaborate_Body);
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));

            N := Process_Request_Spec;
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));
            Bind_AADL_To_Process_Request (Identifier (E), N);

            N := Message_Comment ("Deliver a message to the application");
            Append_Node_To_List (N, ADN.Visible_Part (Current_Package));
         end if;

         Pop_Entity; --  U
         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

   end Package_Spec;

   ------------------
   -- Package_Body --
   ------------------

   package body Package_Body is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);
      procedure Visit_Subprogram_Instance (E : Node_Id);

      Alternatives : List_Id;
      --  The case alternatives of the Process_Request subprogram

      function Process_Request_Body (E : Node_Id) return Node_Id;
      --  Creates the body for the process request operation. E is the
      --  root system instance node.

      function Build_Case_Alternative (E : Node_Id) return Node_Id;
      --  Build the part of the case statement that handle the
      --  execution of subprogram instance E.

      ----------------------------
      -- Build_Case_Alternative --
      ----------------------------

      function Build_Case_Alternative (E : Node_Id) return Node_Id is
         Enumerator    : constant Node_Id := Copy_Node
           (ADN.Enumerator_Node
            (Backend_Node
             (Identifier
              (E))));
         Declarations  : constant List_Id := New_List (ADN.K_Declaration_List);
         Statements    : constant List_Id := New_List (ADN.K_Statement_List);
         Call_Seq      : constant Node_Id := Parent_Sequence
           (Parent_Subcomponent (E));
         Parent_Thread : constant Node_Id := Get_Container_Thread (E);
         Spg_Call     : Node_Id;
         N            : Node_Id;
      begin
         pragma Assert (Is_Thread (Parent_Thread));

         --  Invoke all the subprograms of the call sequence

         Handle_Call_Sequence
           (Parent_Thread, Call_Seq, Declarations, Statements, PN (P_Req));

         --  Trigger all remote threads

         if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then

            Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

            while Present (Spg_Call) loop
               Call_Remote_Subprograms
                 (Parent_Thread, Spg_Call, Declarations, Statements);

               Spg_Call := Next_Node (Spg_Call);
            end loop;
         end if;

         --  Build the block statement

         N := Make_Block_Statement
           (Declarative_Part => Declarations,
            Statements       => Statements);

         --  Build the case alternative

         N := Make_Case_Statement_Alternative
           (Make_List_Id (Enumerator), Make_List_Id (N));
         return N;
      end Build_Case_Alternative;

      --------------------------
      -- Process_Request_Body --
      --------------------------

      function Process_Request_Body (E : Node_Id) return Node_Id is
         Spec         : constant Node_Id := ADN.Process_Request_Node
           (Backend_Node (Identifier (E)));
         Declarations : constant List_Id := New_List (ADN.K_Declaration_List);
         Statements   : constant List_Id := New_List (ADN.K_Statement_List);
         N            : Node_Id;
      begin
         --  Declarations

         N := Make_Subprogram_Call
           (RE (RE_Unmarshall_Request),
            Make_List_Id (Make_Designator (PN (P_Message))));

         N := Make_Object_Declaration
           (Defining_Identifier => Make_Defining_Identifier (PN (P_Req)),
            Constant_Present    => True,
            Object_Definition   => Extract_Designator
              (ADN.Request_Type_Node (Backend_Node (Identifier (E)))),
            Expression          => N);
         Append_Node_To_List (N, Declarations);

         --  Statements

         --  Adding a 'when others => null;' statement to the clase
         --  alternative list to void compiling errors.

         N :=  Make_Case_Statement_Alternative (No_List, No_List);
         Append_Node_To_List (N, Alternatives);

         N := Make_Case_Statement
           (Expression                  => Make_Designator
              (CN (C_Operation),
               PN (P_Req)),
            Case_Statement_Alternatives => Alternatives);
         Append_Node_To_List (N, Statements);

         --  Make the subprogram implementation

         N := Make_Subprogram_Implementation
           (Spec, Declarations, Statements);
         return N;
      end Process_Request_Body;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when CC_Subprogram =>
               Visit_Subprogram_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         --  Visit all the subcomponents of the process

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;
      end Visit_Process_Instance;

      -------------------------------
      -- Visit_Subprogram_Instance --
      -------------------------------

      procedure Visit_Subprogram_Instance (E : Node_Id) is
         N : Node_Id;
      begin
         --  If the subprogram is remote-callable (if it has at least
         --  one IN parameter), generate the corresponding case
         --  alternative in the body of the Process_Request
         --  subprogram.

         if Has_In_Parameters (E) then
            --  Add the case alternative once per subprogram

            if No
              (Get_Handling
               (E,
                By_Name,
                H_Add_Case_Alternative_SSRA_Body))
            then
               N := Build_Case_Alternative (E);
               Set_Handling
                 (E,
                  By_Name,
                  H_Add_Case_Alternative_SSRA_Body,
                  N);
               Append_Node_To_List (N, Alternatives);
            end if;
         end if;
      end Visit_Subprogram_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
         N : Node_Id;
         U : constant Node_Id := ADN.Distributed_Application_Unit
           (ADN.Internals_Node (Backend_Node (Identifier (E))));
      begin
         Push_Entity (HI_Distributed_Application_Root);
         Push_Entity (U);
         Set_SSRA_Body;

         --  Create the case alternative list

         Alternatives := New_List (ADN.K_Variant_List);

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         --  If there is no request type generated for the
         --  application, we generate nothing

         if Present
           (ADN.Request_Type_Node
            (Backend_Node
             (Identifier
              (E))))
         then

            --  Add a use clause to the _Internals package for more
            --  readability

            Add_With_Package
              (Extract_Designator
               (ADN.Internals_Package
                (Current_Entity)),
              Used => True);

            --  Create the body of the Process_Request subprogram

            N := Process_Request_Body (E);
            Append_Node_To_List (N, ADN.Statements (Current_Package));
         end if;

         Pop_Entity; --  U
         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         P        : constant Supported_Thread_Dispatch_Protocol :=
           Get_Thread_Dispatch_Protocol (E);
         Call_Seq : Node_Id;
         Spg_Call : Node_Id;
      begin
         Check_Thread_Consistency (E);

         if P = Thread_Sporadic then
            --  Visit all the call sequences of the thread

            if not AAU.Is_Empty (Calls (E)) then
               Call_Seq := First_Node (Calls (E));

               while Present (Call_Seq) loop
                  --  For each call sequence visit all the called
                  --  subprograms.

                  if not AAU.Is_Empty (Subprogram_Calls (Call_Seq)) then
                     Spg_Call := First_Node (Subprogram_Calls (Call_Seq));

                     while Present (Spg_Call) loop
                        Visit (Corresponding_Instance (Spg_Call));

                        Spg_Call := Next_Node (Spg_Call);
                     end loop;
                  end if;

                  Call_Seq := Next_Node (Call_Seq);
               end loop;
            end if;
         end if;
      end Visit_Thread_Instance;

   end Package_Body;

end Ocarina.Generators.PO_HI_Ada.SSRA;
