------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--  O C A R I N A . G E N E R A T O R S . P O _ H I _ C . A C T I V I T Y   --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                 Copyright (C) 2007, GET-Telecom Paris.                   --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Namet;

with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Generators.Utils;
with Ocarina.Entities.Components;
with Ocarina.Generators.Properties;
with Ocarina.Generators.C_Tree.Nutils;
with Ocarina.Generators.C_Tree.Nodes;
with Ocarina.Generators.PO_HI_C.Mapping;
with Ocarina.Generators.PO_HI_C.Runtime;
with Ocarina.Generators.C_Values;
with Ocarina.Generators.Messages;

package body Ocarina.Generators.PO_HI_C.Activity is

   use Namet;
   use Ocarina.Nodes;
   use Ocarina.Generators.Utils;
   use Ocarina.Entities.Components;
   use Ocarina.Generators.Properties;
   use Ocarina.Generators.C_Tree.Nutils;
   use Ocarina.Generators.PO_HI_C.Mapping;
   use Ocarina.Generators.PO_HI_C.Runtime;
   use Ocarina.Generators.Messages;

   package AAU renames Ocarina.Nutils;
   package CTN renames Ocarina.Generators.C_Tree.Nodes;
   package CTU renames Ocarina.Generators.C_Tree.Nutils;
   package CV renames Ocarina.Generators.C_Values;
   ------------
   -- Header --
   ------------

   package body Header_File is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);
      function Task_Job_Spec (E : Node_Id) return Node_Id;

      -------------------
      -- Task_Job_Spec --
      -------------------

      function Task_Job_Spec (E : Node_Id) return Node_Id is
         N : Node_Id;
         S : constant Node_Id := Parent_Subcomponent (E);
      begin
         N := Make_Function_Specification
           (Defining_Identifier => Map_Task_Job_Identifier (S),
            Parameters          => No_List,
            Return_Type         => CTU.Make_Pointer_Type
              (New_Node (CTN.K_Void))
           );
         return N;
      end Task_Job_Spec;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         U : constant Node_Id := CTN.Distributed_Application_Unit
           (CTN.Naming_Node (Backend_Node (Identifier (E))));
         P : constant Node_Id := CTN.Entity (U);
         S : Node_Id;
      begin
         Push_Entity (P);
         Push_Entity (U);
         Set_Activity_Header (U);
         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;
         Pop_Entity; -- U
         Pop_Entity; -- P
      end Visit_Process_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         Push_Entity (HI_Distributed_Application_Root);

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         P : constant Supported_Thread_Dispatch_Protocol :=
           Get_Thread_Dispatch_Protocol (E);
         S : constant Node_Id := Parent_Subcomponent (E);
         N : Node_Id;
      begin
         --  For each periodic thread, we define the spec of the Job
         --  subprogram. and instantiate a Cyclic_Task.

         case P is
            when Thread_Periodic =>

               N := Message_Comment
                 ("Periodic task : "
                  & Get_Name_String (Display_Name (Identifier (S))));
               Append_Node_To_List (N, CTN.Declarations (Current_File));

               --  Create the spec of the parameterless subprogram that
               --  executes the thread job.

               N := Task_Job_Spec (E);
               Append_Node_To_List (N, CTN.Declarations (Current_File));
               Bind_AADL_To_Job (Identifier (S), N);

               --  Instantiate the cyclic task

               --  N := Cyclic_Task_Instantiation (E);
               --  Append_Node_To_List (N, Declarations (Current_File));

            when Thread_Sporadic =>
               N := Message_Comment
                 ("Sporadic task : "
                  & Get_Name_String (Display_Name (Identifier (S))));
               Append_Node_To_List (N, CTN.Declarations (Current_File));

               N := Task_Job_Spec (E);
               Append_Node_To_List (N, CTN.Declarations (Current_File));
               Bind_AADL_To_Job (Identifier (S), N);

            when others =>
               Display_Error ("unknown thread type", Fatal => True);
         end case;
      end Visit_Thread_Instance;

   end Header_File;

   ------------
   -- Source --
   ------------

   package body Source_File is

      procedure Visit_Architecture_Instance (E : Node_Id);
      procedure Visit_Component_Instance (E : Node_Id);
      procedure Visit_System_Instance (E : Node_Id);
      procedure Visit_Process_Instance (E : Node_Id);
      procedure Visit_Thread_Instance (E : Node_Id);

      function Cyclic_Task_Job_Body (E : Node_Id) return Node_Id;
      function Sporadic_Task_Job_Body (E : Node_Id) return Node_Id;
      --  Creates the parameterless subprogram body that does the
      --  thread's job.

      --  FIXME : In the future, we will have
      --  several system instances. Need to handle all
      --  of them later.
      --      Current_System_Instance : Node_Id := No_Node;

      --  Points to the current root system

      --------------------------
      -- Cyclic_Task_Job_Body --
      --------------------------

      function Cyclic_Task_Job_Body (E : Node_Id) return Node_Id is
         S               : constant Node_Id := Parent_Subcomponent (E);
         Spec            : constant Node_Id := CTN.Job_Node
           (Backend_Node (Identifier (S)));
         Declarations    : constant List_Id := New_List
           (CTN.K_Declaration_List);
         Statements      : constant List_Id := New_List (CTN.K_Statement_List);
         N               : Node_Id;
         WStatements     : constant List_Id := New_List (CTN.K_Statement_List);
         Call_Parameters : constant List_Id := New_List (CTN.K_Parameter_List);
         Call_Seq        : Node_Id;
      begin
         if not AAU.Is_Empty (Calls (E)) then
            Check_Thread_Consistency (E);

            Call_Seq := First_Node (Calls (E));
            CTU.Handle_Call_Sequence (E, Call_Seq, Declarations, WStatements);
         end if;

         N := Make_Defining_Identifier (Map_C_Enumerator_Name (S));
         Append_Node_To_List (N, Call_Parameters);

         N := CTU.Make_Call_Profile (RE (RE_Wait_For_Next_Period),
                                     Call_Parameters);
         Append_Node_To_List (N, WStatements);

         N := Make_While_Statement
           (Make_Literal (CV.New_Int_Value (1, 0, 10)),
            WStatements);
         Append_Node_To_List (N, Statements);
         N := Make_Function_Implementation
           (Spec, Declarations, Statements);
         return N;
      end Cyclic_Task_Job_Body;

      function Sporadic_Task_Job_Body (E : Node_Id) return Node_Id is
         S               : constant Node_Id := Parent_Subcomponent (E);
         Spec            : constant Node_Id := CTN.Job_Node
           (Backend_Node (Identifier (S)));
         Declarations    : constant List_Id := New_List
           (CTN.K_Declaration_List);
         Statements      : constant List_Id := New_List (CTN.K_Statement_List);
         N               : Node_Id;
         WStatements     : constant List_Id := New_List (CTN.K_Statement_List);
         Call_Parameters : List_Id := New_List (CTN.K_Parameter_List);
         Call_Seq        : Node_Id;
      begin
         --  Make __po_hi_sporadic_wait( task_id ) call
         N := Make_Defining_Identifier (Map_C_Enumerator_Name (S));
         Append_Node_To_List (N, Call_Parameters);

         N := CTU.Make_Call_Profile (RE (RE_Sporadic_Wait),
                                     Call_Parameters);
         Append_Node_To_List (N, WStatements);

         if not AAU.Is_Empty (Calls (E)) then
            Check_Thread_Consistency (E);

            Call_Seq := First_Node (Calls (E));
            CTU.Handle_Call_Sequence (E, Call_Seq, Declarations, WStatements);
         end if;

         --  Make __po_hi_sporadic_wait_release( task_id ) call
         Call_Parameters := New_List (CTN.K_Parameter_List);
         N := Make_Defining_Identifier (Map_C_Enumerator_Name (S));
         Append_Node_To_List (N, Call_Parameters);

         N := CTU.Make_Call_Profile (RE (RE_Sporadic_Wait_Release),
                                     Call_Parameters);
         Append_Node_To_List (N, WStatements);

         --  Make the while (1){} and add all statements
         N := Make_While_Statement
           (Make_Literal (CV.New_Int_Value (1, 0, 10)),
            WStatements);
         Append_Node_To_List (N, Statements);

         --  Make the implementation of the function with
         --  all statement defined before
         N := Make_Function_Implementation
           (Spec, Declarations, Statements);
         return N;
      end Sporadic_Task_Job_Body;

      -----------
      -- Visit --
      -----------

      procedure Visit (E : Node_Id) is
      begin
         case Kind (E) is
            when K_Architecture_Instance =>
               Visit_Architecture_Instance (E);

            when K_Component_Instance =>
               Visit_Component_Instance (E);

            when others =>
               null;
         end case;
      end Visit;

      ---------------------------------
      -- Visit_Architecture_Instance --
      ---------------------------------

      procedure Visit_Architecture_Instance (E : Node_Id) is
      begin
         Visit (Root_System (E));
      end Visit_Architecture_Instance;

      ------------------------------
      -- Visit_Component_Instance --
      ------------------------------

      procedure Visit_Component_Instance (E : Node_Id) is
         Cathegory : constant Component_Category
           := Get_Category_Of_Component (E);
      begin
         case Cathegory is
            when CC_System =>
               Visit_System_Instance (E);

            when CC_Process =>
               Visit_Process_Instance (E);

            when CC_Thread =>
               Visit_Thread_Instance (E);

            when others =>
               null;
         end case;
      end Visit_Component_Instance;

      ----------------------------
      -- Visit_Process_Instance --
      ----------------------------

      procedure Visit_Process_Instance (E : Node_Id) is
         U : constant Node_Id := CTN.Distributed_Application_Unit
           (CTN.Naming_Node (Backend_Node (Identifier (E))));
         P : constant Node_Id := CTN.Entity (U);
         S : Node_Id;
      begin
         Push_Entity (P);
         Push_Entity (U);
         CTU.Set_Activity_Source (U);
         --  Visit all the subcomponents of the process
         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; -- U
         Pop_Entity; -- P
      end Visit_Process_Instance;

      ---------------------------
      -- Visit_System_Instance --
      ---------------------------

      procedure Visit_System_Instance (E : Node_Id) is
         S : Node_Id;
      begin
         Push_Entity (HI_Distributed_Application_Root);

         --  FIXME : In the future, we will have
         --  several system instances. Need to handle all
         --  of them later.
         --  Current_System_Instance := E;

         --  Visit all the subcomponents of the system

         if not AAU.Is_Empty (Subcomponents (E)) then
            S := First_Node (Subcomponents (E));
            while Present (S) loop
               --  Visit the component instance corresponding to the
               --  subcomponent S.

               Visit (Corresponding_Instance (S));
               S := Next_Node (S);
            end loop;
         end if;

         Pop_Entity; --  HI_Distributed_Application_Root
      end Visit_System_Instance;

      ---------------------------
      -- Visit_Thread_Instance --
      ---------------------------

      procedure Visit_Thread_Instance (E : Node_Id) is
         P : constant Supported_Thread_Dispatch_Protocol :=
           Get_Thread_Dispatch_Protocol (E);
         S : constant Node_Id := Parent_Subcomponent (E);
         N : Node_Id;
      begin
         --  For each periodic thread, we define the spec of the Job
         --  subprogram. and instantiate a Cyclic_Task.
         case P is
            when Thread_Periodic =>
               N := Message_Comment
                 ("Periodic task : "
                  & Get_Name_String (Display_Name (Identifier (S))));
               Append_Node_To_List (N, CTN.Declarations (Current_File));

               --  Create the body of the parameterless subprogram that
               --  executes the thread job.

               N := Cyclic_Task_Job_Body (E);
               Append_Node_To_List (N, CTN.Declarations (Current_File));
               --  FIXME : Should be bound in the future
               --  Have to check if it needs to be bound
               --  Bind_AADL_To_Job (Identifier (S), N);

            when Thread_Sporadic =>
               N := Message_Comment
                 ("Sporadic task : "
                  & Get_Name_String (Display_Name (Identifier (S))));
               Append_Node_To_List (N, CTN.Declarations (Current_File));

               --  Create the body of the parameterless subprogram that
               --  executes the thread job.

               N := Sporadic_Task_Job_Body (E);
               Append_Node_To_List (N, CTN.Declarations (Current_File));

            when others =>
               Display_Error ("unknown type of thread", Fatal => False);
               null;
         end case;
      end Visit_Thread_Instance;

   end Source_File;

end Ocarina.Generators.PO_HI_C.Activity;
