------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--   O C A R I N A . G E N E R A T O R S . P O _ H I _ C . M A P P I N G    --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                 Copyright (C) 2007, GET-Telecom Paris.                   --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Namet; use Namet;
with Utils; use Utils;
with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Generators.Messages;
with Ocarina.Generators.Utils;
with Ocarina.Generators.C_Values;
with Ocarina.Generators.C_Tree.Nodes;
with Ocarina.Generators.C_Tree.Nutils;
with Ocarina.Generators.PO_HI_C.Runtime;

package body Ocarina.Generators.PO_HI_C.Mapping is

   use Ocarina.Nodes;
   use Ocarina.Generators.Messages;
   use Ocarina.Generators.Utils;
   use Ocarina.Generators.C_Values;
   use Ocarina.Generators.C_Tree.Nodes;
   use Ocarina.Generators.C_Tree.Nutils;
   use Ocarina.Generators.PO_HI_C.Runtime;

   package AAN renames Ocarina.Nodes;
   package AAU renames Ocarina.Nutils;
   package CN renames Ocarina.Generators.C_Tree.Nodes;
   package CU renames Ocarina.Generators.C_Tree.Nutils;

   -----------------------------
   -- Call_Remote_Subprograms --
   -----------------------------

   procedure Call_Remote_Subprograms
     (Caller_Thread    : Node_Id;
      Spg_Call         : Node_Id;
      Declarations     : List_Id;
      Statements       : List_Id)
   is
      procedure Check_Connection_Consistency (F : Node_Id);
      --  Verifies that the feature F is connected to at least one
      --  destination.

      procedure Call_Remote_Subprograms
        (Root_F         : Node_Id;
         Intermediate_F : Node_Id);
      --  This recursive procedure will unwind recursively all the
      --  destinations of the Root_F feature until it finds a
      --  subprogram. If the remote subprogram is found, add a call
      --  to the subprogram call table (if necessary) and add a
      --  parameter association corresponding to Root_F the the
      --  call to this subprogram. At the top level call to
      --  Call_Remote_Subprograms, Intermediate_F = Root_F.

      procedure Update_Remote_Calls
        (Remote_Spg  : Node_Id);
      --  If the call to Spg has not been added yet, create
      --  it. Adds a parameter association Param_Name =>
      --  Param_Value to the call profile.

      Call_List : constant List_Id := New_List (CN.K_Statement_List);
      --  List that contains the calls to all remote subprogram
      --  connected to Spg.

      ----------------------------------
      -- Check_Connection_Consistency --
      ----------------------------------

      procedure Check_Connection_Consistency (F : Node_Id) is
      begin
         if AAU.Length (Destinations (F)) = 0 then
            Display_Located_Error
              (AAN.Loc (F),
               "This feature does not have any destinations",
               Fatal => True);
         end if;
      end Check_Connection_Consistency;

      -----------------------------
      -- Call_Remote_Subprograms --
      -----------------------------

      procedure Call_Remote_Subprograms
        (Root_F         : Node_Id;
         Intermediate_F : Node_Id)
      is
         Destination_F : Node_Id;
         C             : Node_Id;
      begin
         --  Root F has to be a parameter instance

         pragma Assert (Kind (Root_F) = K_Parameter_Instance);

         --  The container of the subprogram containing F has to be
         --  necessarily a Thread.

         pragma Assert (Is_Thread
                        (Parent_Component
                         (Parent_Sequence
                          (Parent_Subcomponent
                           (Parent_Component
                            (Root_F))))));

         --  Check the feature consistency

         Check_Connection_Consistency (Intermediate_F);

         Destination_F := AAN.First_Node (Destinations (Intermediate_F));

         while Present (Destination_F) loop
            C := Parent_Component (Item (Destination_F));

            --  If C is also the parent component of Root_F (which
            --  is necessarily a subprogram). Then display an error
            --  and exit.

            if C = Parent_Component (Root_F) then
               Display_Located_Error
                 (AAN.Loc (C),
                  "This subprogram is involved in a connection cycle",
                  Fatal => True);
            end if;

            if Is_Subprogram (C) then
               --  If C is a subprogram, then append it to the call
               --  list (if necessary, and add a parameter
               --  association corresponding to Root_F and
               --  Destination_F.

               Update_Remote_Calls (C);
            else
               --  Otherwise, keep unwinding the destinations

               Call_Remote_Subprograms (Root_F, Item (Destination_F));
            end if;

            Destination_F := AAN.Next_Node (Destination_F);
         end loop;
      end Call_Remote_Subprograms;

      -------------------------
      -- Update_Remote_Calls --
      -------------------------

      procedure Update_Remote_Calls
        (Remote_Spg  : Node_Id)
      is
         R_Name  : Name_Id;
         N       : Node_Id;
         Profile : List_Id;
      begin
         --  Check whether a call corresponding to Spg has already
         --  been added to Call_List (we use name table infos
         --  instead of looping on Call list, which is a very fast
         --  way.

         Set_Nat_To_Name_Buffer (Nat (Remote_Spg));
         Add_Str_To_Name_Buffer ("%RemoteCall%");
         Add_Nat_To_Name_Buffer (Nat (Spg_Call));
         R_Name := Name_Find;

         if Get_Name_Table_Info (R_Name) = 0 then
            N := Message_Comment ("Call stub ");
            Append_Node_To_List (N, Statements);

            Profile := New_List (CN.K_List_Id);

            --  Create the subprogram call

            N := Make_Call_Profile
              (Identifier
               (Remote_Spg),
               Profile);
            Append_Node_To_List (N, Statements);

            --  Mark the call as being added. The info we associate
            --  to the name is the value of the profile list to be
            --  able to get it to add parameter associations.

            Set_Name_Table_Info (R_Name, Nat (Profile));
         else
            Profile := List_Id (Get_Name_Table_Info (R_Name));
         end if;

      end Update_Remote_Calls;

      Spg              : constant Node_Id := Corresponding_Instance (Spg_Call);
      F                : Node_Id;
   begin
      pragma Assert (Is_Thread (Caller_Thread));

      --  The lists have to be created

      if Declarations = No_List or else Statements = No_List then
         raise Program_Error with "Lists have to be created before any "
           & "call to Get_Remote_Subprogram";
      end if;

      if not AAU.Is_Empty (Features (Spg)) then
         F := AAN.First_Node (Features (Spg));

         while Present (F) loop
            if Kind (F) = K_Parameter_Instance and then Is_Out (F) then
               --  Call all the remote subprograms connected the
               --  feature F.

               Call_Remote_Subprograms (F, F);
            end if;

            F := AAN.Next_Node (F);
         end loop;
      end if;

      --  Append the calls to the Statements list

      if not CU.Is_Empty (Call_List) then
         Append_Node_To_List (CN.First_Node (Call_List), Statements);

      end if;
   end Call_Remote_Subprograms;

   ---------------------------------
   -- Map_Distributed_Application --
   ---------------------------------

   function Map_Distributed_Application (E : Node_Id) return Node_Id is
      D : constant Node_Id := New_Node (CN.K_HI_Distributed_Application);
   begin
      pragma Assert (Is_System (E));

      --  Update the global variable to be able to fetch the root of
      --  the distributed application and generate the source files.

      HI_Distributed_Application_Root := D;

      CN.Set_Name (D, To_C_Name
                   (AAN.Name
                    (AAN.Identifier
                     (E))));
      CN.Set_Units (D, New_List (CN.K_List_Id));
      CN.Set_HI_Nodes (D, New_List (CN.K_List_Id));

      return D;
   end Map_Distributed_Application;

   -----------------
   -- Map_HI_Node --
   -----------------

   function Map_HI_Node (E : Node_Id) return Node_Id is
      N : constant Node_Id := New_Node (CN.K_HI_Node);
   begin
      pragma Assert (Is_Process (E));

      --  The name of the node is not the name of the process
      --  component instance, but the name of the process subcomponent
      --  corresponding to this instance.

      CN.Set_Name (N, To_C_Name
                    (AAN.Name
                     (AAN.Identifier
                      (AAN.Parent_Subcomponent
                       (E)))));

      Set_Units (N, New_List (K_List_Id));

      --  Append the partition N to the node list of the PolyORB-HI
      --  distributed application. We are sure that the top of the
      --  entity stack contains the Ada distrubuted application node.

      Append_Node_To_List (N, HI_Nodes (Current_Entity));
      Set_Distributed_Application (N, Current_Entity);

      return N;
   end Map_HI_Node;

   -----------------
   -- Map_HI_Unit --
   -----------------

   function Map_HI_Unit (E : Node_Id) return Node_Id is
      U        : Node_Id;
      S        : List_Id;
      H        : List_Id;
      N        : Node_Id;
      P        : Node_Id;
   begin
      pragma Assert (Is_System (E) or else Is_Process (E));

      U := New_Node (CN.K_HI_Unit, AAN.Identifier (E));
      S := New_List (K_Sources);
      H := New_List (K_Headers);

      --  Packages that are common to all nodes

      if Is_Process (E) then

         Set_Str_To_Name_Buffer ("activity");
         N := Make_Defining_Identifier (Name_Find);
         P := Make_Source_File (N);
         Set_Distributed_Application_Unit (P, U);
         CN.Set_Activity_Source (U, P);
         Append_Node_To_List (P, S);
         Bind_AADL_To_Activity (Identifier (E), P);

         Set_Str_To_Name_Buffer ("naming");
         N := Make_Defining_Identifier (Name_Find);
         P := Make_Source_File (N);
         Set_Distributed_Application_Unit (P, U);
         CN.Set_Naming_Source (U, P);
         Append_Node_To_List (P, S);
         Bind_AADL_To_Naming (Identifier (E), P);

         Set_Str_To_Name_Buffer ("activity");
         N := Make_Defining_Identifier (Name_Find);
         P := Make_Header_File (N);
         Set_Distributed_Application_Unit (P, U);
         CN.Set_Activity_Header (U, P);
         Append_Node_To_List (P, H);

         Set_Str_To_Name_Buffer ("main");
         N := Make_Defining_Identifier (Name_Find);
         P := Make_Source_File (N);
         Set_Distributed_Application_Unit (P, U);
         CN.Set_Main_Source (U, P);
         Append_Node_To_List (P, S);
         Bind_AADL_To_Main (Identifier (E), P);

         Set_Str_To_Name_Buffer ("deployment");
         N := Make_Defining_Identifier (Name_Find);
         P := Make_Header_File (N);
         Set_Distributed_Application_Unit (P, U);
         CN.Set_Deployment_Header (U, P);
         Append_Node_To_List (P, H);

--         Bind_AADL_To_Activity (Identifier (E), P);

--         P := Make_Main_Function_Implementation
--           (Make_Defining_Identifier (C_Name));
--         Set_Distributed_Application_Unit (P, U);
--         Set_Main_Function (U, P);
--         Append_Node_To_List (P, S);
--         Bind_AADL_To_Main (Identifier (E), P);
      end if;

      --  Append the Unit to the units list of the current Ada
      --  partition.

      CN.Set_Sources (U, S);
      CN.Set_Headers (U, H);

      Append_Node_To_List (U, Units (Current_Entity));
      CN.Set_Entity (U, Current_Entity);

      return U;
   end Map_HI_Unit;

   ---------------------------
   -- Bind_AADL_To_Activity --
   ---------------------------

   procedure Bind_AADL_To_Activity (G : Node_Id; A : Node_Id) is
      N : Node_Id;
   begin
      N := AAN.Backend_Node (G);

      if No (N) then
         N := New_Node (CN.K_HI_Tree_Bindings);
         AAN.Set_Backend_Node (G, N);
      end if;

      CN.Set_Activity_Node (N, A);
      CN.Set_Frontend_Node (A, G);
   end Bind_AADL_To_Activity;

   -------------------------
   -- Bind_AADL_To_Naming --
   -------------------------

   procedure Bind_AADL_To_Naming (G : Node_Id; A : Node_Id) is
      N : Node_Id;
   begin
      N := AAN.Backend_Node (G);

      if No (N) then
         N := New_Node (CN.K_HI_Tree_Bindings);
         AAN.Set_Backend_Node (G, N);
      end if;

      CN.Set_Naming_Node (N, A);
      CN.Set_Frontend_Node (A, G);
   end Bind_AADL_To_Naming;

   ----------------------
   -- Bind_AADL_To_Job --
   ----------------------

   procedure Bind_AADL_To_Job (G : Node_Id; A : Node_Id) is
      N : Node_Id;
   begin
      N := AAN.Backend_Node (G);

      if No (N) then
         N := New_Node (CN.K_HI_Tree_Bindings);
         AAN.Set_Backend_Node (G, N);
      end if;

      CN.Set_Job_Node (N, A);
      CN.Set_Frontend_Node (A, G);
   end Bind_AADL_To_Job;

   -----------------------------
   -- Bind_AADL_To_Enumerator --
   -----------------------------

   procedure Bind_AADL_To_Enumerator (G : Node_Id; A : Node_Id) is
      N : Node_Id;
   begin
      N := AAN.Backend_Node (G);

      if No (N) then
         N := New_Node (CN.K_HI_Tree_Bindings);
         AAN.Set_Backend_Node (G, N);
      end if;

      CN.Set_Enumerator_Node (N, A);
      CN.Set_Frontend_Node (A, G);
   end Bind_AADL_To_Enumerator;

   -----------------------
   -- Bind_AADL_To_Main --
   -----------------------

   procedure Bind_AADL_To_Main (G : Node_Id; A : Node_Id) is
      N : Node_Id;
   begin
      N := AAN.Backend_Node (G);

      if No (N) then
         N := New_Node (CN.K_HI_Tree_Bindings);
         AAN.Set_Backend_Node (G, N);
      end if;

      CN.Set_Main_Node (N, A);
      CN.Set_Frontend_Node (A, G);
   end Bind_AADL_To_Main;

   -----------------------------
   -- Map_Task_Job_Identifier --
   -----------------------------

   function Map_Task_Job_Identifier (E : Node_Id) return Node_Id is
      Name : Name_Id;
   begin
      Get_Name_String (To_C_Name (AAN.Display_Name (Identifier (E))));
      Add_Str_To_Name_Buffer ("_job");
      Name := Name_Find;
      Name := To_Lower (Name);
      return Make_Defining_Identifier (Name);
   end Map_Task_Job_Identifier;

   --------------
   -- Map_Time --
   --------------

   function Map_Time (T : Time_Type) return Node_Id is
      Time : Unsigned_Long_Long;
      S    : Node_Id;
   begin
      case T.U is
         when Picosecond =>
            --  Our framework only support microseconds
            --  Picosecond and Nanosecond are ignored
            if T.T mod 1000000 = 0 then
               Time := T.T / 1000000;
               S := RE (RE_Microseconds);
            else
               return No_Node;
            end if;

         when Nanosecond =>
            if T.T mod 1000 = 0 then
               Time := T.T / 1000;
               S := RE (RE_Microseconds);
            else
               return No_Node;
            end if;
            S := RE (RE_Microseconds);

         when Microsecond =>
            Time := T.T;
            S := RE (RE_Microseconds);

         when Millisecond =>
            Time := T.T;
            S := RE (RE_Milliseconds);

         when Second =>
            Time := T.T;
            S := RE (RE_Seconds);

         when Minute =>
            Time := T.T;
            S := RE (RE_Minutes);

         when Hour =>
            --  Convert it into minutes

            Time := T.T * 60;
            S := RE (RE_Minutes);
      end case;

      return Make_Call_Profile
        (S, Make_List_Id (Make_Literal (New_Int_Value (Time, 1, 10))));
   end Map_Time;

   ---------------------------
   -- Map_C_Enumerator_Name --
   ---------------------------

   function Map_C_Enumerator_Name
     (E      : Node_Id;
      Server : Boolean := False)
     return Name_Id
   is
      Ada_Name_1 : Name_Id;
      Ada_Name_2 : Name_Id;
   begin
      pragma Assert
        (Is_Subprogram (E) or else Kind (E) = K_Subcomponent_Instance);

      if Is_Subprogram (E)
        or else Is_Process (Corresponding_Instance (E))
      then
         --  For subprograms and processes, the enemerator name is
         --  mapped from the entity name.

         Get_Name_String (CU.To_C_Name (Display_Name (Identifier (E))));
         Add_Str_To_Name_Buffer ("_k");
      elsif Is_Thread (Corresponding_Instance (E)) then
         --  For threads, the enumerator name is mapped from the
         --  containing process name and the thread subcomponent name.

         --  Verifiy that the thread is a subcomponent of a process

         pragma Assert (Is_Process (Parent_Component (E)));

         Ada_Name_1 := CU.To_C_Name
           (Display_Name
            (Identifier
             (Parent_Subcomponent
              (Parent_Component
               (E)))));
         Ada_Name_2 := CU.To_C_Name
           (Display_Name
            (Identifier
             (E)));

         Get_Name_String (Ada_Name_1);
         Add_Char_To_Name_Buffer ('_');
         Get_Name_String_And_Append (Ada_Name_2);
         Add_Str_To_Name_Buffer ("_k");
      else
         raise Program_Error with
           "Wrong node kind for Map_C_Enumerator_Name";
      end if;

      if Server then
         Add_Str_To_Name_Buffer ("_Server");
      end if;

      return Name_Find;
   end Map_C_Enumerator_Name;

end Ocarina.Generators.PO_HI_C.Mapping;
