------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--                             O C A R I N A 1                              --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2004-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

--  This program is used to drive Ocarina, it is a wrapper to all
--  functions provided by the library.

with Output;
with Namet;
with GNAT.OS_Lib;
with Ocarina_Options;
with Ocarina.Parser;
with Ocarina.Printer;
with Ocarina.Expander;
with Types;
with Ocarina;
with Ocarina.Configuration;
with Ocarina.Analyzer;
with Ocarina.Generators;
with Gaia.Mgmt;
with Ocarina.PN;
with Ocarina.Checker;
with Ocarina.Checker_Testsuite;

procedure Ocarina1 is
   use Output;
   use Types;
   use Namet;
   use Ocarina.Configuration;
   use Ocarina.Analyzer;
   use Ocarina.Generators;
   use Ocarina.Parser;
   use Ocarina.Printer;
   use Ocarina.Expander;
   use Ocarina.PN;
   use Gaia.Mgmt;
   use GNAT.OS_Lib;
   use Ocarina_Options;
   use Ocarina_Options.Source_File_Pkg;
   use Ocarina.Checker;
   use Ocarina.Checker_Testsuite;

   Root                   : Node_Id;
   Success                : Boolean := True;
   Files                  : Source_File_List;
   Action                 : Operation;
   Generator_Options      : Gaia_Options;
   Generators_Options     : Ocarina.Generators.Generator_Options;
   Analysis_Options       : Analyzer_Options;
   Printer_Options        : Output_Options;
   Petri_Net_Options      : PN_Options;
   Checker_Options        : Ocarina.Checker.Checker_Options;
   Do_Parse_Predefined_PS : Boolean;
begin
   --  Initialization

   Ocarina.Initialize;
   Ocarina.Configuration.Init_Modules;

   --  Configuration

   Generator_Options  := Default_Gaia_Options;
   Generators_Options := Default_Options;
   Analysis_Options   := Default_Analyzer_Options;
   Printer_Options    := Default_Output_Options;
   Petri_Net_Options  := Default_PN_Options;
   Checker_Options    := Ocarina.Checker.Default_Checker_Options;

   --  Process the command line

   Process_Command_Line
     (Action,
      Files,
      Generator_Options,
      Generators_Options,
      Analysis_Options,
      Printer_Options,
      Petri_Net_Options,
      Checker_Options,
      Do_Parse_Predefined_PS,
      Success);

   if Success then
      case Action is
         when Version =>
            Write_Line ("Ocarina "
                        & Ocarina_Version
                        & " ("
                        & Ocarina_SVN_Revision
                        & ")");

            if Ocarina_Last_Configure_Date /= "" then
               Write_Line ("Build date: " & Ocarina_Last_Configure_Date);
            end if;

            Write_Line
              ("Copyright (c) 2003-"
               & Ocarina_Last_Configure_Year
               & " ENST");

            Success := True;

         when Search_Dir =>
            Write_Line (Ocarina.Configuration.Data_Directory);
            Success := True;

         when others =>
            --  Initialize thr AADL tree root to No_Node to signal a
            --  new parsing.

            Root := No_Node;

            --  Parsing the AADL specification

            if Success and then Last (Files) /= 0 then
               for F_Index in First .. Last (Files) loop
                  Root := Parse
                    (Get_Name_String (Files.Table (F_Index)), Root);
                  Success := Present (Root);
               end loop;

               if No (Root) then
                  Set_Standard_Error;
                  Write_Line ("Error while parsing the files");
               end if;
            end if;

            --  After parsing all the user given files, try to parse
            --  the standard property files.

            if Success then
               Root := Parse_Standard_Property_Sets (Root);
               Success := Present (Root);
            end if;

            --  If the user requested explicitely to parse the Ocarina
            --  predefined property sets, do it. Otherwise, assume he
            --  gave their corresponding files to the command line.

            if Success and then Do_Parse_Predefined_PS then
               Root := Parse_Ocarina_Property_Sets (Root);
               Success := Present (Root);

               if not Success then
                  Set_Standard_Error;
                  Write_Line ("Unable to parse Ocarina predefined"
                              & " property sets");
               end if;
            end if;

            --  Analysis

            if Success then
               Success := Analyze_Tree (Root, Analysis_Options);

               if not Success then
                  Set_Standard_Error;
                  Write_Line ("Error when analyzing the tree");
               end if;
            end if;

            --  Action

            if Success then

               if Checker_Options.Perform_Checking then
                  Perform_All_Checking (Root, Checker_Options);
               end if;

               case Action is
                  when Do_Print =>
                     Success := Print (Root, Printer_Options);

                     if not Success then
                        Write_Line ("Printer not found");
                     end if;

                  when Expand =>
                     declare
                        Instance_Root : constant Node_Id :=
                          Expand_Model (Root);
                     begin
                        if No (Instance_Root) then
                           Write_Line ("The AADL model cannot be expanded.");
                           Success := False;
                        end if;
                     end;

                  when Old_Petri_Net =>
                     Success := Generate (Root, Generator_Options);

                     Set_Standard_Error;

                     if Success then
                        Write_Line ("Petri net successfully generated");
                     else
                        Write_Line ("Petri net generatation failed");
                     end if;

                  when Generate_Code =>
                     Generate_Code (Root, Generators_Options);

                     --  If we get to this point, this meands the code
                     --  generation has been successful.

                     Success := True;

                  when Dump_Gen_Tree =>
                     Generator_Options.Dump_Gaia_Tree := True;
                     Success := Generate (Root, Generator_Options);

                     Set_Standard_Error;

                     if Success then
                        Write_Line ("Gaia tree successfully dumped");
                     else
                        Write_Line ("Gaia tree dumping failed");
                     end if;

                  when Petri_Net =>
                     Success := Create_Petri_Net (Root, Petri_Net_Options);

                     Set_Standard_Error;

                     if Success then
                        Write_Line ("Petri Net successfully generated");
                     else
                        Write_Line ("Petri Net generatation failed");
                     end if;

                  when others =>
                     null;

               end case;
            end if;
      end case;
   end if;

   if Success then
      OS_Exit (0);
   else
      OS_Exit (1);
   end if;
end Ocarina1;
