------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--                      O C A R I N A _ O P T I O N S                       --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2006-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with GNAT.OS_Lib;
with GNAT.Command_Line;
with Ada.Command_Line;
with Output;
with Namet;
with Gaia.Utils;
with Ocarina.Parser;
with Ocarina.Generators.Messages;

package body Ocarina_Options is

   type Criteria is (Nothing, No_Null_Length);
   --  Criteria to check parameter validity

   function Checked_Parameter
     (Parameter : String;
      Criterion : Criteria)
     return String;
   --  Checks the validity of a parameter and returns it in case it's valid

   -----------------------
   -- Checked_Parameter --
   -----------------------

   function Checked_Parameter
     (Parameter : String;
      Criterion : Criteria)
     return String
   is
   begin
      case Criterion is

         when No_Null_Length =>
            if Parameter'Length > 0 then
               return Parameter;
            end if;

         when Nothing =>
            return Parameter;
      end case;

      return "";
   end Checked_Parameter;

   --------------------------
   -- Process_Command_Line --
   --------------------------

   procedure Process_Command_Line
     (Action            :    out Operation;
      Files             : in out Source_File_List;
      Gaia_Options      : in out Gaia.Mgmt.Gaia_Options;
      Generator_Options : in out Ocarina.Generators.Generator_Options;
      Analysis_Options  : in out Ocarina.Analyzer.Analyzer_Options;
      Printer_Options   : in out Ocarina.Printer.Output_Options;
      Petri_Net_Options : in out Ocarina.PN.PN_Options;
      Checker_Options   : in out Ocarina.Checker.Checker_Options;
      Parse_PPS         :    out Boolean;
      Success           :    out Boolean)
   is
      use Namet;
      use Gaia.Mgmt;
      use Gaia.Utils;
      use Ocarina.Parser;
      use Ocarina.Analyzer;
      use GNAT.Command_Line;
      use GNAT.OS_Lib;
      use Source_File_Pkg;
   begin
      Success := True;
      Parse_PPS := False;
      Init (Files);
      Action := Do_Nothing;

      Initialize_Option_Scan;

      loop
         case Getopt ("b o: d: g: I: p: n c f u e q a v V s h") is
            when 'b' =>
               Generator_Options.Build := True;

            when 'f' =>
               Parse_PPS := True;

            when 'o' =>
               Printer_Options.Output_File := Get_String_Name
                 (Checked_Parameter (Parameter, No_Null_Length));

            when 'd' =>
               declare
                  Output_Dir : constant Name_Id := Get_String_Name
                    (Checked_Parameter (Parameter, No_Null_Length));
               begin
                  Gaia_Options.Output_Directory      := Output_Dir;
                  Generator_Options.Output_Directory := Output_Dir;
                  Printer_Options.Output_Directory   := Output_Dir;
                  Petri_Net_Options.Output_Directory := Output_Dir;
               end;

            when 'g' =>
               Action := Generate_Code;
               Ocarina.Generators.Select_Generator
                 (Checked_Parameter
                  (Parameter, No_Null_Length));

            when 'I' =>
               Add_Search_Path (Checked_Parameter (Parameter, No_Null_Length));

            when 'p' =>
               Action := Do_Print;
               Printer_Options.Printer_Name := Get_String_Name
                 (Checked_Parameter (Parameter, No_Null_Length));

            when 'c' =>
               Action := Old_Petri_Net;
               Gaia_Options.Generator_Name  := Get_String_Name ("ps");
               Gaia_Options.Translator_Name := Get_String_Name ("ps");

            when 'u' =>
               Action := Dump_Gen_Tree;

            when 'n' =>
               Action := Petri_Net;

            when 'e' =>
               Action := Expand;

            when 'a' =>
               Analysis_Options.Legality_Rules := Devel;

            when 'q' =>
               Analysis_Options.Display_Debug_Messages := False;
               Ocarina.Generators.Messages.Output_Debug_Messages := False;

            when 'V' =>
               Analysis_Options.Display_Debug_Messages := True;
               Ocarina.Generators.Messages.Output_Debug_Messages := True;

            when 'v' =>
               Action := Version;
               exit;

            when 's' =>
               Action := Search_Dir;
               exit;

            when 'h' =>
               Checker_Options.Perform_Checking := True;
               Checker_Options.Check_Mutexes := True;
               exit;

            when ASCII.NUL =>
               exit;

            when others =>
               Usage;
         end case;
      end loop;

      loop
         Set_Str_To_Name_Buffer (Get_Argument);

         if Name_Len = 0 then
            exit;
         else
            Append (Files, Name_Find);
         end if;
      end loop;

      --  If no file is given, print the usage function and exit

      if Last (Files) = 0
        and then Action /= Version
        and then Action /= Search_Dir
      then
         Usage;
      end if;
   exception
      when GNAT.Command_Line.Invalid_Switch => Usage;
   end Process_Command_Line;

   -----------
   -- Usage --
   -----------

   procedure Usage is
      use Output;
      use GNAT.OS_Lib;
      use Ada.Command_Line;
      use Ocarina.Printer;
   begin
      Set_Standard_Error;
      Write_Str  ("Usage: ");
      Write_Str  (Command_Name);
      Write_Line (" [opts] files");
      Write_Eol;
      Write_Line ("  files are a non null sequence of AADL or DIA files");
      Write_Eol;
      Write_Line ("  Options:");
      Write_Line ("  -v  Output Ocarina version, then exit");
      Write_Line ("  -s  Output Ocarina search directory, then exit");
      Write_Line ("  -f  Parse Ocarina predefined NON STANDARD property sets");
      Write_Line ("  -a  Alternative legality rules. Extensions to AADL 1.0");
      Write_Line ("  -c  To generate Petri Net from an intermediary tree");
      Write_Line ("  -n  To generate Petri Net from the AADL instance model");
      Write_Line ("  -g  To generate code from the AADL instance tree");
      Ocarina.Generators.Write_Generators (6);
      Write_Line ("  -u  Dump the gaia tree");
      Write_Line ("  -e  To only expand the AADL model");
      Write_Line ("  -d  Specify output directory");
      Write_Line ("  -o  Specify output file");
      Write_Line ("  -p  Specify the printer to use");
      Write_Printers (6);
      Write_Line ("  -I  Specify the inclusion paths");
      Write_Line ("  -q  Quiet mode, no debugging messages (default mode)");
      Write_Line ("  -V  Verbose mode, display debugging messages");
      Write_Line ("  -h  Hardware checking, perform hardware checking");
      Write_Eol;
      OS_Exit (1);
   end Usage;

end Ocarina_Options;
