------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--                    O C A R I N A _ S C E N A R I O S                     --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                 Copyright (C) 2007, GET-Telecom Paris.                   --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Strings.Unbounded;

with GNAT.Table;
with GNAT.Directory_Operations;

with Namet;
with Output;
with Types;

with Ocarina.Nodes;
with Ocarina.Nutils;
with Ocarina.Parser;
with Ocarina.Analyzer;
with Ocarina.Expander.Queries;
with Ocarina.AADL_Values;

with Ocarina.Configuration;
--  FIXME: This is totally a bad thing because we drag all the ocarina
--  libraries. The right thing is to use setup packages to register
--  the ocarina modules.

package body Ocarina_Scenarios is

   use Ada.Strings.Unbounded;
   use GNAT.Directory_Operations;
   use Namet;
   use Output;
   use Types;
   use Ocarina.Nodes;
   use Ocarina.Nutils;
   use Ocarina.Parser;
   use Ocarina.Analyzer;
   use Ocarina.Expander;
   use Ocarina.Expander.Queries;
   use Ocarina.AADL_Values;
   use Ocarina.Configuration;

   package Current_AADL_File_List is new
     GNAT.Table (Name_Id, Natural, 1, 5, 100);
   --  The current list of AADL files

   Current_Scenario_Dirname : Unbounded_String := To_Unbounded_String ("");
   --  The current dirname of all the scenario files.

   Current_Generator : Name_Id := No_Name;
   --  The current code generator

   procedure Display_Error (Message : String);
   --  Display and error message and exit the program with an erro
   --  status of 2.

   procedure Extract_Source_Files
     (Source_Files : List_Id;
      Needed_PS    : List_Id);
   --  Update the Current_AADL_File_List with the full path of the
   --  given AADL source file list an the needed predefined property
   --  sets.

   First_Scenario : Boolean := True;
   --  Used to control the Reset/Init of the Ocarina engines

   Ocarina_Config : constant String := "ocarina_config";
   --  The property set containing the Ocarina configuration
   --  properties.

   AADL_Files    : constant String := Ocarina_Config & "::aadl_files";
   The_Generator : constant String := Ocarina_Config & "::generator";
   Predefined_PS : constant String := Ocarina_Config
     & "::needed_property_sets";
   --  Some miscellaneous properties

   -------------------
   -- Display_Error --
   -------------------

   procedure Display_Error (Message : String) is
   begin
      Set_Standard_Error;
      Write_Line (Message);
      OS_Exit (2);
   end Display_Error;

   --------------------------
   -- Extract_Source_Files --
   --------------------------

   procedure Extract_Source_Files
     (Source_Files : List_Id;
      Needed_PS    : List_Id)
   is
      use Current_AADL_File_List;

      N : Node_Id;
   begin
      N := First_Node (Source_Files);

      while Present (N) loop
         Set_Str_To_Name_Buffer (To_String (Current_Scenario_Dirname));

         if Name_Buffer (Name_Len) /= Directory_Separator then
            Add_Char_To_Name_Buffer (Directory_Separator);
         end if;

         Get_Name_String_And_Append (Value (Value (N)).SVal);

         Append (Name_Find);

         N := Next_Node (N);
      end loop;

      if not Is_Empty (Needed_PS) then
         N := First_Node (Needed_Ps);

         while Present (N) loop
            declare
               P : Name_Id;
            begin
               Get_Name_String (Value (Value (N)).EVal);
               Set_Str_To_Name_Buffer
                 (Image
                  (Ocarina_Property_Sets'Value
                   ("O_" & Name_Buffer (1 .. Name_Len))));
               P := Name_Find;

               Set_Str_To_Name_Buffer (Data_Directory & Directory_Separator);
               Get_Name_String_And_Append (P);
               Add_Str_To_Name_Buffer (".aadl");
            exception
               when others =>
                  Display_Error ("Unknown predefined property set "
                                 & Get_Name_String (Value (Value (N)).SVal));
            end;

            Append (Name_Find);

            N := Next_Node (N);
         end loop;
      end if;
   end Extract_Source_Files;

   -------------------------
   -- Parse_Scenario_File --
   -------------------------

   procedure Parse_Scenario_File (Scenario : Argument_List_Access) is
      AADL_Root      : Node_Id := No_Node;
      Instance_Root  : Node_Id := No_Node;
      Root_System    : Node_Id := No_Node;
      Source_Files   : List_Id;
      Needed_PS      : List_Id;
      Temp_Dirname   : Unbounded_String;
      Success        : Boolean := False;
   begin
      --  Reset Ocarina if this is not the first parsed scenario to
      --  allow further scenario files parsing.

      if not First_Scenario then
         Ocarina.Configuration.Reset_Modules;
         Ocarina.Reset;
         First_Scenario := False;
      end if;

      --  Initialize a light version of the Ocarina libraries that
      --  allows just to parse and expand an AADL file.

      Ocarina.Initialize;
      Ocarina.Configuration.Init_Modules;
      Current_AADL_File_List.Free;
      Current_AADL_File_List.Init;
      Current_Scenario_Dirname := To_Unbounded_String ("");

      --  Parse the AADL scenario files

      for J in Scenario'Range loop
         Temp_Dirname := To_Unbounded_String
           (Normalize_Pathname
            (Dir_Name
             (Scenario (J).all)));
         AADL_Root := Parse (Scenario (J).all, AADL_Root);

         Success := Present (AADL_Root) and then
           (Current_Scenario_Dirname = To_Unbounded_String ("") or else
            Current_Scenario_Dirname = Temp_Dirname);
         exit when not Success;

         if Current_Scenario_Dirname = To_Unbounded_String ("") then
            Current_Scenario_Dirname := Temp_Dirname;
         end if;
      end loop;

      --  Parse the Ocarina_Config property set. No need to parse the
      --  AADL standard property set since a scenario is not suposed
      --  to use them.

      if Success then
         AADL_Root := Search_And_Parse (Ocarina_Config, AADL_Root, False);
         Success := Present (AADL_Root);
      end if;

      --  Check that the parsing went properly

      if not Success then
         Display_Error ("*** Error while parsing the scenario files."
                        & " Scenario files must be located in the same"
                        & " directory ***");
      end if;

      --  Analyze the AADL tree

      Success := Analyze_Tree (AADL_Root, Default_Analyzer_Options);

      if not Success then
         Display_Error ("*** Error while analyzing the scenario files ***");
      end if;

      --  Expand the AADL tree

      Instance_Root := Expand_Model (AADL_Root);
      Success := Present (Instance_Root);

      if not Success then
         Display_Error ("*** Error while expanding the scenario files ***");
      end if;

      --  Every thing is fine, extract the information from the
      --  SCENARIO model.

      Root_System := Ocarina.Nodes.Root_System (Instance_Root);

      --  Extract the generator

      if Is_Defined_Enumeration_Property (Root_System, The_Generator) then
         Current_Generator := Get_Enumeration_Property
           (Root_System, The_Generator);
      else
         Display_Error ("You must specify a code generator");
      end if;

      --  Extract the AADL files to be parsed by the Compiler

      if Is_Defined_List_Property (Root_System, AADL_Files) then
         Source_Files := Get_List_Property (Root_System, AADL_Files);
      else
         Display_Error ("AADL source files have to be specified by"
                        & " means of the standard property"
                        & " ""Source_Text""");
      end if;

      --  Extract the Preefined property sets needed by the
      --  application

      if Is_Defined_List_Property (Root_System, Predefined_PS) then
         Needed_PS := Get_List_Property (Root_System, Predefined_PS);
      else
         Needed_PS := No_List;
      end if;

      --  Normalize the paths to all files in order to give them to
      --  Ocarina.

      Extract_Source_Files (Source_Files, Needed_PS);
   end Parse_Scenario_File;

   ----------------------
   -- Get_Source_Files --
   ----------------------

   function Get_Source_Files return Argument_List_Access is
      use Current_AADL_File_List;

      Result : constant Argument_List_Access :=
        new Argument_List (First .. Last);
   begin
      for J in First .. Last loop
         Result (J) := new String'(Get_Name_String (Table (J)));
      end loop;

      return Result;
   end Get_Source_Files;

   -------------------
   -- Get_Generator --
   -------------------

   function Get_Generator return String is
   begin
      return Get_Name_String (Current_Generator);
   end Get_Generator;

end Ocarina_Scenarios;
