------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--                    O C A R I N A _ S C E N A R I O S                     --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--                 Copyright (C) 2007, GET-Telecom Paris.                   --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

--  This package contains the routines that parse the AADL scenario
--  files and deduce information that will be passed to the Ocarina
--  compiler.

--  An AADL scenario file is a file (or more) containing an AADL
--  system component *implementation*. The properties of this
--  systems will let Ocarina deduce the following information:

--  1 - the list of AADL source files used to describe the AADL
--  application.

--  2 - the action that has to be performed on these files (pasing,
--  code generation, schedulability analysis...).

with GNAT.OS_Lib; use GNAT.OS_Lib;

package Ocarina_Scenarios is

   procedure Parse_Scenario_File (Scenario : Argument_List_Access);
   --  Parses and AADL scenario files and raises several error
   --  messages corresponding to the possible missing, wrong or
   --  duplicated imformation in this file. NOTE all the given files
   --  must contains only one unique scenario (see the example
   --  below). For the sake of simplicity, all the scenario files MUST
   --  be located in the same directory when parsing them.

   function Get_Source_Files return Argument_List_Access;
   --  Return a table of String_Access each one representing the full
   --  path to an AADL file that has to be parsed by the Ocarina
   --  compiler. The returned list corresponds to the latest parsed
   --  AADL scenario file. The construction of the argument list
   --  requires some memory allocation on the heap. The deallocation
   --  of this memory is up to the caller of Get_Source_Files. This
   --  returned list includes all the Ocarina predefined NON  STANDARD
   --  property sets needed by the application as well.

   function Get_Generator return String;
   --  Return a lower case string representing the action to be done
   --  by the compiler. The returned action corresponds to the latest
   --  parsed AADL scenario file. We intentionaly do not use an
   --  enumeration to represent all the possible actions in order to
   --  allo users to define new actions and new treatment for
   --  them. All the associations between the actions and the
   --  treatment correspodning to them should be summarized in a file
   --  similar the 'specs' file of the gcc compiler. The action
   --  nature will be deduced from the AADL property attributed to
   --  the scenario system.

   --  Example:

   --  -----------------
   --  -- FILE_1.aadl --
   --  -----------------

   --  system RMA
   --  properties
   --    Ocarina_Config::AADL_Files => ("rma.aadl");
   --    --  "rma.aadl" contains common AADL components (processes,
   --    --  threads, data types)
   --
   --    Ocarina_Config::Needed_Property_Sets => (ARAO, Cheddar_Properties);
   --    --  The non standard predefined property sets needed by the
   --    --  application.
   --  end RMA;

   --  -----------------
   --  -- FILE_2.aadl --
   --  -----------------

   --  system implementation RMA.Impl_C
   --  properties
   --    Ocarina_Config::AADL_Files +=> ("software_c.aadl");
   --    --  Note that this is an additive property
   --    --  association.
   --
   --    Ocarina_Config::Generator => PolyORB_HI_C;
   --    --  The code generator
   --  end RMA.Impl_C;

   --  -----------------
   --  -- FILE_3.aadl --
   --  -----------------

   --  system implementation RMA.Impl_Ada
   --  properties
   --    Source_Text +=> ("software_ada.aadl");
   --    --  Note that this is an additive property
   --    --  association.
   --
   --    Ocarina_Config::Generator => PolyORB_HI_Ada;
   --    --  The code generator
   --  end RMA.Impl_Ada;

   --  Calling the compiler on FILE_1.aadl and FILE_2.aadl will
   --  generate C code for the Poly_ORB-HI framework starting from
   --  "rma.aadl" and "software_c.aadl".

   --  Calling the compiler on FILE_1.aadl and FILE_3.aadl will
   --  generate Ada code for the Poly_ORB-HI framework starting from
   --  "rma.aadl" and "software_ada.aadl".

end Ocarina_Scenarios;
