------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--                           O C A R I N A _ S H                            --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                 Copyright (C) 2007, GET-Telecom Paris.                   --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Strings.Unbounded;
with Ada.Text_IO;
with Ada.Command_Line;
with GNAT.Command_Line;
with GNAT.Current_Exception;
with GNAT.Expect;
with GNAT.Directory_Operations;
with GNAT.OS_Lib;
with GNAT.Case_Util;

with Ocarina_Scenarios;

procedure Ocarina_Sh is

   use Ada.Strings.Unbounded;
   use Ada.Text_IO;
   use GNAT.Command_Line;
   use GNAT.Current_Exception;
   use GNAT.OS_Lib;
   use GNAT.Expect;
   use GNAT.Directory_Operations;
   use GNAT.Case_Util;
   use Ocarina_Scenarios;

   function Get_Install_Dir (Exec : String) return String;
   --  Extract the install directory from the string Exec. If the
   --  commandline has been invoked without a directory prefix (using
   --  the $PATH or %Path% variable), return "".

   procedure Usage;

   procedure Parse_Command_Line;
   procedure Run_Command
     (Command_Name  :     String;
      Argument_List :     GNAT.OS_Lib.Argument_List;
      Status        : out Integer);
   procedure Run_Print_Version (Status : out Integer);
   procedure Run_Generate_PolyORB_QoS_Ada (Status : out Integer);
   procedure Run_Generate_PolyORB_HI_Ada (Status : out Integer);
   procedure Run_Generate_PolyORB_HI_C (Status : out Integer);
   procedure Run_Produce_PetriScript (Status : out Integer);
   procedure Run_Ocarina_Parser (Status : out Integer);
   procedure Run_Ocarina_Checker (Status : out Integer);
   procedure Run_Cheddar_Analyzer (Status : out Integer);

   Executable_Suffix : String renames Get_Executable_Suffix.all;
   Build             : Boolean := False;
   Print_Version     : Boolean := False;
   Cheddar_Only      : Boolean := False;
   Check_Only        : Boolean := False;
   Parse_Only        : Boolean := False;
   PetriScript_Only  : Boolean := False;
   Null_String       : constant Unbounded_String := To_Unbounded_String ("");
   Example_Dirname   : Unbounded_String := Null_String;
   AADL_Files        : Argument_List_Access := new Argument_List (1 .. 0);

   Ocarina_1        : Unbounded_String;
   Cheddar_Analyzer : Unbounded_String;
   --  Full path to the ocarina1 executable

   ---------------------
   -- Get_Install_Dir --
   ---------------------

   function Get_Install_Dir (Exec : String) return String is
      Skept_Executable_Name : Boolean := False;
   begin
      for J in reverse Exec'Range loop
         if Exec (J) = Directory_Separator then
            if J < Exec'Last - 5 then
               if To_Lower (Exec (J + 1)) = 'b'
                 and then To_Lower (Exec (J + 2)) = 'i'
                 and then To_Lower (Exec (J + 3)) = 'n'
               then
                  return Exec (Exec'First .. J + 3) & Directory_Separator;
               elsif not Skept_Executable_Name then
                  Skept_Executable_Name := True;
               else
                  exit;
               end if;
            end if;
         end if;
      end loop;

      return "";
   end Get_Install_Dir;

   -----------
   -- Usage --
   -----------

   procedure Usage is
      use Ada.Command_Line;
   begin
      Put_Line ("Usage: " & Command_Name
                & " [-v] [-p] [-c] [-n] [-h]"
                & " [-s <scenario_part_1> <scenario_part_2>...]");
      Put_Line (" scenario_part_? is the AADL file LIST describing the"
                & " application scenario");
      Put_Line (" -b: build the generated application code");
      Put_Line (" -p: only parse the application model");
      Put_Line (" -n: only produce Petri Net");
      Put_Line (" -c: only perform schedulability analysis");
      Put_Line (" -h: only perform checks on the application model");
      Put_Line (" -v: output Ocarina version");
   end Usage;

   ------------------------
   -- Parse_Command_Line --
   ------------------------

   procedure Parse_Command_Line is
      Scenario_File : Unbounded_String;
   begin
      Initialize_Option_Scan ('-', False, "s");

      --  Get arguments

      loop
         case Getopt ("b c h n p v") is
            when ASCII.Nul =>
               exit;

            when 'b' =>
               Build := True;

            when 'c' =>
               Cheddar_Only := True;

            when 'h' =>
               Check_Only := True;

            when 'n' =>
               PetriScript_Only := True;

            when 'p' =>
               Parse_Only := True;

            when 'v' =>
               Print_Version := True;

            when others =>
               Usage;
               OS_Exit (1);
         end case;
      end loop;

      --  Get the scenario files

      Goto_Section ("s");

      Scenario_File := To_Unbounded_String (Get_Argument);

      while Scenario_File /= Null_String loop
         declare
            Temp : constant Argument_List_Access
              := new Argument_List (1 .. AADL_Files.all'Length + 1);
         begin
            Temp (1 .. Temp'Last - 1) := AADL_Files.all;
            Temp (Temp'Last) := new String'(To_String (Scenario_File));
            Free (AADL_Files);
            AADL_Files := Temp;
         end;

         Scenario_File := To_Unbounded_String (Get_Argument);
      end loop;

      if AADL_Files'Length = 0 and then not Print_Version then
         Usage;
         OS_Exit (1);
      end if;

   exception
      when GNAT.Command_Line.Invalid_Switch =>
         begin
            Put_Line ("Invalid Switch " & Full_Switch);
            Usage;
            OS_Exit (1);
         end;

      when GNAT.Command_Line.Invalid_Parameter =>
         begin
            Put_Line ("Missing parameter for switch " & Full_Switch);
            OS_Exit (1);
         end;
   end Parse_Command_Line;

   -----------------
   -- Run_Command --
   -----------------

   procedure Run_Command
     (Command_Name  :     String;
      Argument_List :     GNAT.OS_Lib.Argument_List;
      Status        : out Integer)
   is
      Fd     : Process_Descriptor;
      Result : Expect_Match;
      Path   : GNAT.OS_Lib.String_Access := Getenv ("PATH");
   begin
      Status := 0;

      --  Some debugging stuff

      pragma Debug (Put_Line ("Executing command:"));
      pragma Debug (Put_Line (" Name: " & Command_Name & Executable_Suffix));
      pragma Debug (Put      (" Args: "));
      for I in Argument_List'Range loop
         pragma Debug (Put (Argument_List (I).all & " "));
         null;
      end loop;
      pragma Debug (New_Line);
      pragma Debug (Put_Line ("Current directory: " & Get_Current_Dir));
      pragma Debug (Put_Line ("PATH=" & Path.all));
      Free (Path);

      Non_Blocking_Spawn
        (Descriptor  => Fd,
         Command     => Command_Name & Executable_Suffix,
         Args        => Argument_List,
         Buffer_Size => 4096,
         Err_To_Out  => True);

      Add_Filter (Fd, Trace_Filter'Access, GNAT.Expect.Output);
      Add_Filter (Fd, Trace_Filter'Access, GNAT.Expect.Died);

      --  Wait until the command achieves its execution

      begin
         Expect (Fd, Result, "", -1);

         --  This call never match, it is only used to wait until
         --  Ocarina1 terminates.

      exception
         when GNAT.Expect.Process_Died =>
            Close (Fd, Status);

            if Status /= 0 then
               pragma Debug (Put_Line (Command_Name
                                       & Executable_Suffix
                                       & " died unexpectedly with code: "
                                       & Status'Img));
               null;
            end if;
      end;
   end Run_Command;

   -----------------------
   -- Run_Print_Version --
   -----------------------

   procedure Run_Print_Version (Status : out Integer) is
   begin
      Run_Command
        (Command_Name  => To_String (Ocarina_1),
         Argument_List => (1 => new String'("-v")),
         Status        => Status);
   end Run_Print_Version;

   ----------------------------------
   -- Run_Generate_PolyORB_QoS_Ada --
   ----------------------------------

   procedure Run_Generate_PolyORB_QoS_Ada (Status : out Integer) is
      Args : Argument_List_Access;
      Pos  : Natural := 1;
   begin
      if Build then
         Args := new Argument_List (1 .. AADL_Files.all'Length + 3);
         Args (Pos) := new String'("-b");
         Pos := Pos + 1;
      else
         Args := new Argument_List (1 .. AADL_Files.all'Length + 2);
      end if;

      Args (Pos) := new String'("-g");
      Args (Pos + 1) := new String'("PolyORB-QoS-Ada");

      for J in AADL_Files'Range loop
         Args (Pos + 2 - AADL_Files'First + J) :=
           new String'(AADL_Files (J).all);
      end loop;

      Run_Command
        (Command_Name  => To_String (Ocarina_1),
         Argument_List => Args.all,
         Status        => Status);

      Free (Args);
   end Run_Generate_PolyORB_QoS_Ada;

   ---------------------------------
   -- Run_Generate_PolyORB_HI_Ada --
   ---------------------------------

   procedure Run_Generate_PolyORB_HI_Ada (Status : out Integer) is
      Args : Argument_List_Access;
      Pos  : Natural := 1;
   begin
      if Build then
         Args := new Argument_List (1 .. AADL_Files.all'Length + 3);
         Args (Pos) := new String'("-b");
         Pos := Pos + 1;
      else
         Args := new Argument_List (1 .. AADL_Files.all'Length + 2);
      end if;

      Args (Pos) := new String'("-g");
      Args (Pos + 1) := new String'("PolyORB-HI-Ada");

      for J in AADL_Files'Range loop
         Args (Pos + 2 - AADL_Files'First + J) :=
           new String'(AADL_Files (J).all);
      end loop;

      Run_Command
        (Command_Name  => To_String (Ocarina_1),
         Argument_List => Args.all,
         Status        => Status);

      Free (Args);
   end Run_Generate_PolyORB_HI_Ada;

   -------------------------------
   -- Run_Generate_PolyORB_HI_C --
   -------------------------------

   procedure Run_Generate_PolyORB_HI_C (Status : out Integer) is
      Args : Argument_List_Access;
      Pos  : Natural := 1;
   begin
      if Build then
         Args := new Argument_List (1 .. AADL_Files.all'Length + 3);
         Args (Pos) := new String'("-b");
         Pos := Pos + 1;
      else
         Args := new Argument_List (1 .. AADL_Files.all'Length + 2);
      end if;

      Args (Pos) := new String'("-g");
      Args (Pos + 1) := new String'("PolyORB-HI-C");

      for J in AADL_Files'Range loop
         Args (Pos + 2 - AADL_Files'First + J) :=
           new String'(AADL_Files (J).all);
      end loop;

      Run_Command
        (Command_Name  => To_String (Ocarina_1),
         Argument_List => Args.all,
         Status        => Status);

      Free (Args);
   end Run_Generate_PolyORB_HI_C;

   ------------------------
   -- Run_Ocarina_Parser --
   ------------------------

   procedure Run_Ocarina_Parser (Status : out Integer) is
      Args : Argument_List (1 .. AADL_Files.all'Length + 1);
   begin
      Args (1) := new String'("-paadl");

      Args (2 .. Args'Last) := AADL_Files.all;

      Run_Command
        (Command_Name  => To_String (Ocarina_1),
         Argument_List => Args,
         Status        => Status);
   end Run_Ocarina_Parser;

   -------------------------
   -- Run_Ocarina_Checker --
   -------------------------

   procedure Run_Ocarina_Checker (Status : out Integer) is
      Args : Argument_List (1 .. AADL_Files.all'Length + 1);
   begin
      Args (1) := new String'("-h");

      Args (2 .. Args'Last) := AADL_Files.all;

      Run_Command
        (Command_Name  => To_String (Ocarina_1),
         Argument_List => Args,
         Status        => Status);
   end Run_Ocarina_Checker;

   --------------------------
   -- Run_Cheddar_Analyzer --
   --------------------------

   procedure Run_Cheddar_Analyzer (Status : out Integer) is
      Args : constant Argument_List := AADL_Files.all;
   begin
      if To_String (Cheddar_Analyzer) = "" then
         raise Program_Error with "cheddar_analyzer: command not found!";
      end if;

      Run_Command
        (Command_Name  => To_String (Cheddar_Analyzer),
         Argument_List => Args,
         Status        => Status);
   end Run_Cheddar_Analyzer;

   -----------------------------
   -- Run_Produce_PetriScript --
   -----------------------------

   procedure Run_Produce_PetriScript (Status : out Integer) is
      Args        : Argument_List (1 .. AADL_Files.all'Length + 1);
      Current_Dir : constant Dir_Name_Str := Get_Current_Dir;

   begin
      Args (1) := new String'("-n");
      Args (2 .. Args'Last) := AADL_Files.all;

      if not Is_Directory (To_String (Example_Dirname)) then
         Make_Dir (To_String (Example_Dirname));
      end if;

      Change_Dir (To_String (Example_Dirname));

      Run_Command
        (Command_Name  => To_String (Ocarina_1),
         Argument_List => Args,
         Status        => Status);

      --  FIXME: Implement Ada routines equivalent of the scripts that
      --  merge the implementation files with the generated files.

      Change_Dir (Current_Dir);
   end Run_Produce_PetriScript;

   --  Main processing of Ocarina_Sh begins here

   Status : Integer := 0;

   PO_QoS_Ada_Action     : constant String := "polyorb_qos_ada";
   PO_HI_Ada_Action      : constant String := "polyorb_hi_ada";
   PO_HI_C_Action        : constant String := "polyorb_hi_c";
   --  FIXME: The use of these constant should be replaced by a spec
   --  file that allows to execute the correspodning action.

begin
   --  Get the path to the ocarina1 and cheddar_analyzer (if
   --  available) executables which are assumed to be at the same
   --  directory as ocarina_sh.

   declare
      Normalized_Path : constant String := Normalize_Pathname
        (Ada.Command_Line.Command_Name);
      Exec_Dir        : constant String := Get_Install_Dir
        (Normalized_Path);

      O_Basename : constant String := "ocarina";
      --  FIXME: to be changed into ocarina1

      C_Basename : constant String := "cheddar_analyzer";
   begin
      if Exec_Dir /= "" and then Is_Regular_File (Normalized_Path) then
         Ocarina_1 := To_Unbounded_String
           (Normalize_Pathname (Exec_Dir & O_Basename));
         Cheddar_Analyzer := To_Unbounded_String
           (Normalize_Pathname (Exec_Dir & C_Basename));
      else
         --  This means the user called ocarina_sh without directory
         --  prefix. We search our executable in the PATH environment
         --  variable.

         declare
            O : GNAT.OS_Lib.String_Access := Locate_Exec_On_Path (O_Basename);
            C : GNAT.OS_Lib.String_Access := Locate_Exec_On_Path (C_Basename);
         begin
            if O /= null then
               Ocarina_1 := To_Unbounded_String (Normalize_Pathname (O.all));
               Free (O);
            else
               raise Program_Error with "O_Basename: comman not found!";
            end if;

            if C /= null then
               Cheddar_Analyzer := To_Unbounded_String
                 (Normalize_Pathname (C.all));
               Free (C);
            else
               Cheddar_Analyzer := To_Unbounded_String ("");
            end if;
         end;
      end if;
   end;

   Parse_Command_Line;

   if Print_Version then
      Run_Print_Version (Status);
   elsif AADL_Files'Length /= 0 then
      Parse_Scenario_File (AADL_Files);

      --  Free the scenario files list and replace it with the AADL
      --  source files.

      Free (AADL_Files);
      AADL_Files := Get_Source_Files;

      if Parse_Only then
         Run_Ocarina_Parser (Status);
      elsif Check_Only then
         Run_Ocarina_Checker (Status);
      elsif Cheddar_Only then
         Run_Cheddar_Analyzer (Status);
      elsif PetriScript_Only then
         Run_Produce_PetriScript (Status);
      else
         --  Get the wanted Generator

         declare
            Action : constant String := Get_Generator;
         begin
            if Action = PO_QoS_Ada_Action then
               Run_Generate_PolyORB_QoS_Ada (Status);
            elsif Action = PO_HI_Ada_Action then
               Run_Generate_PolyORB_HI_Ada (Status);
            elsif Action = PO_HI_C_Action then
               Run_Generate_PolyORB_HI_C (Status);
            else
               Free (AADL_Files);
               Put_Line ("Unrecognized action: " & Action);
               OS_Exit (2);
            end if;
         end;
      end if;
   end if;

   Free (AADL_Files);
   OS_Exit (Status);
exception
   when others =>
      Put_Line ("Exception name : " & Exception_Name);
      Put_Line ("Exception message : " & Exception_Information);
      OS_Exit (1);
end Ocarina_Sh;
